/*  Kvaser Linux Canlib LAPcan specific functions */
/*  Copyright (C) 2003 Kvaser AB - support@kvaser.com - www.kvaser.com  */

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <canlib.h>

#include <lapcan_data.h>
#include <lapcan_ioctl.h>

#include "canlib_util.h"

#include <sys/types.h>
#include <sys/stat.h>

#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <lapcmds.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>
#include <sys/stat.h>

#include "LapcanFunctions.h"

static unsigned char globalTransId;

void * lpcNotifyThread (void * arg)
{
    HandleData *hData;
    canNotifyData *notifyDataPtr;
    int ret;
    unsigned long time;
    unsigned char cmdNr;
    lpcCmd reply;
    hData = (HandleData*) arg;
    /* Get time to start with */
    hData -> hwOps ->readTimer(hData, &time);
    time *= LAPCAN_TICKS_PER_MS;
    while (1){
        /* Allow cancellation here */ 
        pthread_testcancel();
        ret = ioctl(hData -> notifyFd, LAPCAN_IOC_RECVCMD, &reply);
        /* When this thread is cancelled ioctl will be interrupted by a signal */
        if (ret != 0) pthread_exit(NULL);
        notifyDataPtr = &(hData -> notifyData);
        cmdNr = getCmdNr(&reply);
    
        switch(cmdNr){
        case CMD_CLOCK_OVERFLOW_EVENT:
            time = reply.clockOverflowEvent.currentTime;	
            break;
        case CMD_CHIP_STATE_EVENT:
            notifyDataPtr -> eventType = canEVENT_STATUS;
            notifyDataPtr -> info.status.txErrorCounter = reply.chipStateEvent.txErrorCounter;
            notifyDataPtr -> info.status.rxErrorCounter = reply.chipStateEvent.rxErrorCounter;
            notifyDataPtr -> info.status.time = time & 0xFFFF0000;
            notifyDataPtr -> info.status.time += reply.chipStateEvent.time;
            /* Rescale */
            notifyDataPtr -> info.status.time /= LAPCAN_TICKS_PER_MS;
            notifyDataPtr -> info.status.busStatus = reply.chipStateEvent.busStatus;
            hData -> callback(notifyDataPtr);
            break;
        case CMD_RX_STD_MESSAGE:	
            if (getFlags(&reply) & MSGFLAG_ERROR_FRAME){
                notifyDataPtr -> eventType = canEVENT_ERROR;
                notifyDataPtr -> info.busErr.time = time & 0xFFFF0000;
                notifyDataPtr -> info.busErr.time += reply.rxStdMessage.time;
                notifyDataPtr -> info.busErr.time /= LAPCAN_TICKS_PER_MS;
            } else if (getFlags(&reply) & MSGFLAG_TX){
                notifyDataPtr -> eventType = canEVENT_TX;
                notifyDataPtr -> info.tx.id = reply.rxStdMessage.id;
                notifyDataPtr -> info.tx.time = time & 0xFFFF0000;
                notifyDataPtr -> info.tx.time += reply.rxStdMessage.time;
                notifyDataPtr -> info.tx.time /= LAPCAN_TICKS_PER_MS;
            } else {
                notifyDataPtr -> eventType = canEVENT_RX;
                notifyDataPtr -> info.rx.id = reply.rxStdMessage.id;
                notifyDataPtr -> info.rx.time = time & 0xFFFF0000;
                notifyDataPtr -> info.rx.time += reply.rxStdMessage.time;
                notifyDataPtr -> info.rx.time /= LAPCAN_TICKS_PER_MS;
            }
            hData -> callback(notifyDataPtr);
            break;
        case CMD_RX_EXT_MESSAGE:
            if (getFlags(&reply) & MSGFLAG_ERROR_FRAME){
                notifyDataPtr -> eventType = canEVENT_ERROR;
                notifyDataPtr -> info.busErr.time = time & 0xFFFF0000;
                notifyDataPtr -> info.busErr.time += reply.rxExtMessage.time;
                notifyDataPtr -> info.busErr.time /= LAPCAN_TICKS_PER_MS;
            } else if (getFlags(&reply) & MSGFLAG_TX){
                notifyDataPtr -> eventType = canEVENT_TX;
                notifyDataPtr -> info.tx.id = reply.rxExtMessage.id;
                notifyDataPtr -> info.tx.time = time & 0xFFFF0000;
                notifyDataPtr -> info.tx.time += reply.rxExtMessage.time;
                notifyDataPtr -> info.tx.time /= LAPCAN_TICKS_PER_MS;
            } else {
                notifyDataPtr -> eventType = canEVENT_RX;
                notifyDataPtr -> info.rx.id = reply.rxExtMessage.id;
                notifyDataPtr -> info.rx.time = time & 0xFFFF0000;
                notifyDataPtr -> info.rx.time += reply.rxExtMessage.time;
                notifyDataPtr -> info.rx.time /= LAPCAN_TICKS_PER_MS;
            }
            hData -> callback(notifyDataPtr);
            break;
        }
    }
}


canStatus lpcSetNotify (HandleData *hData, 
                        void (*callback) (canNotifyData *), unsigned int notifyFlags)
{
    int ret;

    if (hData -> notifyFd == 0) {
        /* Open a "file" to receive notification from */
        hData -> notifyFd = open(hData -> deviceName, O_RDONLY);
        if (hData -> notifyFd == -1) goto error_nomem;
        ret = pthread_create(&(hData -> notifyThread),NULL, lpcNotifyThread, hData);
        if (ret != 0) goto error_nomem;
    }

    /* Set filters */
    clearFilters(hData -> notifyFd);
    ret = addFilter(hData -> notifyFd, CMD_CLOCK_OVERFLOW_EVENT,-1, 
                    0, 0, 0, 0, 1);
    if (ret != 0) goto error_nomem;

    if (notifyFlags & canNOTIFY_RX) {
        /* Read rx messages from our channel STD, EXT but not echoed messages */
        ret = addFilter(hData -> notifyFd, CMD_RX_STD_MESSAGE, 0xFFFE, 
                        hData -> channelNr, -1, 0, MSGFLAG_TX | MSGFLAG_TXRQ, 1);
        if (ret != 0) goto error_nomem;
    }
    if (notifyFlags & canNOTIFY_TX) {
        /* Read echoed rx messages from our channel STD, EXT with MSGFLAG_TX set */
        ret = addFilter(hData -> notifyFd, CMD_RX_STD_MESSAGE, 0xFFFE, 
                        hData -> channelNr, -1, MSGFLAG_TX, MSGFLAG_TX, 1);
        if (ret != 0) goto error_nomem;
    }
    if (notifyFlags & canNOTIFY_STATUS) {
        /* Read rx messages from our channel STD, EXT */
        ret = addFilter(hData -> notifyFd, CMD_CHIP_STATE_EVENT, -1, 
                        hData -> channelNr, -1, 0, 0, 1);
        if (ret != 0) goto error_nomem;
    }
    if (notifyFlags & canNOTIFY_ERROR) {
        /* Read rx messages from our channel STD, EXT with errorframe set */
        ret = addFilter(hData -> notifyFd, CMD_RX_STD_MESSAGE, 0xFFFE, 
                        hData -> channelNr, -1, 
                        MSGFLAG_ERROR_FRAME, MSGFLAG_ERROR_FRAME, 1);
        if (ret != 0) goto error_nomem;
    }
    /* Block everything else */
    ret = addFilter(hData -> notifyFd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto error_nomem;

    ioctl(hData -> notifyFd ,LAPCAN_IOC_FLUSH_RCVBUFFER, 0);  
    hData -> callback = callback;
    return canOK;

 error_nomem:
    close(hData -> notifyFd);
    return canERR_NOMEM;
}

canStatus lpcOpenChannel (HandleData *hData)
{
    hData->fd = open(hData->deviceName, O_RDONLY);
    if (hData->fd == -1) goto file_error;  
    if (ioctl(hData->fd, LAPCAN_IOC_OPEN_CHAN, &hData->channelNr)) goto file_error;
    /* LAPCAN_IOC_OPEN_CHAN sets channelNr to -1 if it fails */
    if (hData->channelNr < 0) goto channel_error;
    if (hData->wantExclusive) {
        if (ioctl(hData->fd, LAPCAN_IOC_SET_EXCL, &hData->channelNr)) goto file_error; 
        if (hData->channelNr < 0) goto channel_error;
    }
    /* Read rx messages & from our channel STD & EXT but no transmit request replys */
    if (addFilter(hData->fd, CMD_RX_STD_MESSAGE, 0xFFFE, hData->channelNr, -1, 0, MSGFLAG_TX, 1)) goto filter_error;
    /* We also want CMD_CLOCK_OVERFLOW_EVENT */
    if (addFilter(hData->fd, CMD_CLOCK_OVERFLOW_EVENT, -1, 0, 0, 0, 0, 1)) goto filter_error;
    /* Block everything else */
    if (addFilter(hData->fd, 0, 0, 0, 0, 0, 0, 0)) goto filter_error;

    return canOK;
  
 filter_error:
    close(hData->fd);
    return canERR_NOMEM;
 file_error:
    close(hData->fd);
    return canERR_NOTFOUND;
 channel_error:
    close(hData->fd);
    return canERR_NOCHANNELS;  
}

canStatus lpcBusOn (HandleData *hData) {
    int fd;
    int ret;
    cmdStartChipReq cmd;
    cmdStartChipResp reply;
    fd = open(hData -> deviceName, O_RDONLY);

    /* Block everything but CMD_START_CHIP_RESP */
    ret = addFilter(fd, CMD_START_CHIP_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo = CMD_START_CHIP_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);
    if (ret != 0) goto ioctl_error;
    while(1) {
        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) goto ioctl_error;
        if (reply.transId == cmd.transId){
            close(fd);
            return canOK;
        }
    }
 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
    return canERR_INVHANDLE;
}

canStatus lpcBusOff (HandleData *hData)
{
    int fd;
    int ret;
    cmdStopChipReq cmd;
    cmdStopChipResp reply;

    fd = open(hData -> deviceName, O_RDONLY);
    /* Block everything but CMD_STOP_CHIP_RESP */
    ret = addFilter(fd, CMD_STOP_CHIP_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd,LAPCAN_IOC_FLUSH_RCVBUFFER, 0);

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo = CMD_STOP_CHIP_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    ret = ioctl(fd, LAPCAN_IOC_SENDCMD, &cmd);
    if (ret != 0) goto ioctl_error;

    while(1) {
        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) goto ioctl_error;
        if (reply.transId == cmd.transId){
            close(fd);
            return canOK;
        }
    }
 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
  
    return canERR_INVHANDLE;
}
canStatus lpcSetBusParams (HandleData *hData, long freq, unsigned int tseg1,
                           unsigned int tseg2, unsigned int sjw, unsigned int noSamp,
                           unsigned int syncmode)
{
    int ret;
    cmdSetBusparamsReq cmd;

    cmd.cmdLen  = CMDLEN(cmd);
    cmd.cmdNo   = CMD_SET_BUSPARAMS_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    cmd.bitRate = (unsigned long) freq;
    cmd.tseg1   = (unsigned char) tseg1;
    cmd.tseg2   = (unsigned char) tseg2;
    cmd.sjw     = (unsigned char) sjw;
    cmd.noSamp  = (unsigned char) noSamp;
    /* Syncmode ??? */
    do {
        ret = ioctl(hData -> fd, LAPCAN_IOC_SENDCMD, &cmd);
    } while (ret != 0 && errno == EAGAIN);
    if (ret != 0) return canERR_PARAM;
	   
    return canOK;
}
canStatus lpcGetBusParams(HandleData *hData, long * freq, unsigned int * tseg1,
                          unsigned int * tseg2, unsigned int * sjw,
                          unsigned int * noSamp, unsigned int * syncmode)
{
    cmdGetBusparamsReq cmd;
    cmdGetBusparamsResp reply;
    int fd;
    int ret;

    fd = open(hData -> deviceName, O_RDONLY);
    /* Block everything but CMD_GET_BUSPARAMS_RESP */
    ret = addFilter(fd, CMD_GET_BUSPARAMS_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo = CMD_GET_BUSPARAMS_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);

    while(1) {
        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) goto ioctl_error;
        if (reply.transId == cmd.transId) break;
    }

    close(fd);
    *freq = reply.bitRate;
    *tseg1 = reply.tseg1;
    *tseg2 = reply.tseg2;
    *sjw   = reply.sjw;
    *noSamp = reply.noSamp;
    *syncmode = 0;
    /* Syncmode ??? */
    return canOK;

 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
    return canERR_INVHANDLE;
}
   
static canStatus lpcReadInternal (HandleData * hData, long * id, 
                                  void * msgPtr, unsigned int * dlc, 
                                  unsigned int * flag, unsigned long * time)
{
    int i;
    int tmpDlc;
    lpcCmd tmpMsg;

    while (1) {
        if (ioctl(hData -> fd ,LAPCAN_IOC_RECVCMD, &tmpMsg) != 0){
            if (errno == EAGAIN) return canERR_NOMSG;
            else if (errno == EINTR) return canERR_INTERRUPTED;
            else return canERR_INVHANDLE;
        }
        if (getCmdNr(&tmpMsg) == CMD_RX_STD_MESSAGE) {
            /* Receive standard message */
            tmpDlc =  tmpMsg.rxStdMessage.channel_and_dlc & 0x0F;  /* Bit 0..3 is dlc */
            if (id != NULL) *id = tmpMsg.rxStdMessage.id;
            if (dlc != NULL) *dlc = tmpDlc;
            if (time != NULL) {
                *time = (volatile long)hData -> currentTime;
                *time &= 0xFFFF0000; /* Clear lower 16 bits */
                *time += tmpMsg.rxStdMessage.time;
                *time /= LAPCAN_TICKS_PER_MS;
            }
            if (flag != NULL) {
                *flag = canMSG_STD;      
                if (tmpMsg.rxStdMessage.flags & MSGFLAG_OVERRUN) *flag |=canMSGERR_HW_OVERRUN | canMSGERR_SW_OVERRUN;;
                if (tmpMsg.rxStdMessage.flags & MSGFLAG_REMOTE_FRAME) *flag |= canMSG_RTR;
                if (tmpMsg.rxStdMessage.flags & MSGFLAG_ERROR_FRAME) *flag |= canMSG_ERROR_FRAME;
            }
            if (msgPtr != NULL) {
                for (i = 0; i < tmpDlc; i++) /* Copy data */
                    ((unsigned char*) msgPtr) [i] = tmpMsg.rxStdMessage.data [i];
            }
            break;
        } else if (getCmdNr(&tmpMsg) == CMD_RX_EXT_MESSAGE){
            /* Receive extended message */
            tmpDlc = tmpMsg.rxExtMessage.channel_and_dlc & 0x0F;
            if (id != NULL) *id = tmpMsg.rxExtMessage.id;
            if (dlc != NULL) *dlc =tmpDlc;
            if (time != NULL) {
                *time = (volatile long) hData -> currentTime;
                *time &= 0xFFFF0000; /* Clear lower 16 bits */
                *time += tmpMsg.rxExtMessage.time;
                *time /= LAPCAN_TICKS_PER_MS;
            }
            if (flag != NULL) {
                *flag = canMSG_EXT;
                if (tmpMsg.rxExtMessage.flags & MSGFLAG_OVERRUN) *flag |= canMSGERR_HW_OVERRUN | canMSGERR_SW_OVERRUN;
                if (tmpMsg.rxExtMessage.flags & MSGFLAG_REMOTE_FRAME) *flag |= canMSG_RTR;
                if (tmpMsg.rxExtMessage.flags & MSGFLAG_ERROR_FRAME) *flag |= canMSG_ERROR_FRAME;
            }
            if (msgPtr != NULL) {
                for (i = 0; i < *dlc; i++) /* Copy data */
                    ((unsigned char*) msgPtr) [i] = tmpMsg.rxExtMessage.data [i];
            }
            break;
        } else if (getCmdNr(&tmpMsg) == CMD_CLOCK_OVERFLOW_EVENT) {
            hData -> currentTime = tmpMsg.clockOverflowEvent.currentTime;
            continue;
        } 
    }
    return canOK;
}

canStatus lpcRead (HandleData * hData, long * id, void * msgPtr, 
                   unsigned int * dlc, unsigned int * flag, unsigned long * time)
{
    /* Set non blocking fileop */
    if (hData -> readIsBlock) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_READ_BLOCK, 0);
    hData -> readIsBlock = 0;
    return lpcReadInternal(hData, id, msgPtr, dlc, flag, time);
}

canStatus lpcReadWait (HandleData * hData, long * id, 
                       void * msgPtr, unsigned int * dlc, unsigned int * flag, 
                       unsigned long * time, long timeout)
{
    if (timeout == 0) return lpcRead(hData, id, msgPtr, dlc, flag, time);
    /* Set blocking fileop */
    if (!hData -> readIsBlock) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_READ_BLOCK, 1);
    /* Set timeout in milliseconds */
    if (hData -> readTimeout != timeout) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_READ_TIMEOUT, timeout);
    hData -> readIsBlock = 1;
    hData -> readTimeout = timeout;
    return lpcReadInternal(hData, id, msgPtr, dlc, flag, time);
}

canStatus lpcSetBusOutputControl (HandleData * hData, unsigned int drivertype)
{
    cmdSetDrivermodeReq cmd;
    int ret;

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo  = CMD_SET_DRIVERMODE_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    switch (drivertype) {
    case canDRIVER_NORMAL: 
        cmd.driverMode = DRIVERMODE_NORMAL;
        break;
    case canDRIVER_SILENT: 
        cmd.driverMode = DRIVERMODE_SILENT;
        break;
    case canDRIVER_SELFRECEPTION: 
        return canERR_PARAM; /* ??? NOT IMPLEMENTED ??? */
        cmd.driverMode = DRIVERMODE_SELFRECEPTION;
        break;
    case canDRIVER_OFF: 
        cmd.driverMode = DRIVERMODE_OFF;
        break;
    }
    do {
        ret = ioctl(hData -> fd, LAPCAN_IOC_SENDCMD, &cmd);
    } while (ret != 0 && errno == EAGAIN);
    if (ret != 0) return canERR_INVHANDLE;
    return canOK;
}
canStatus lpcGetBusOutputControl (HandleData * hData, unsigned int * drivertype)
{
    cmdGetDrivermodeReq cmd;
    cmdGetDrivermodeResp reply;
    int fd;
    int ret;

    fd = open(hData -> deviceName, O_RDONLY);

    /* Block everything but CMD_GET_DRIVERMODE_RESP */
    ret = addFilter(fd, CMD_GET_DRIVERMODE_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo  = CMD_GET_DRIVERMODE_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr;
    ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);

    while(1) {
        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) return canERR_NOTFOUND;
        if (reply.transId == cmd.transId) break;
    }
 
    close(fd);
    switch (reply.driverMode) {
    case DRIVERMODE_NORMAL: 
        *drivertype = canDRIVER_NORMAL;
        break;
    case DRIVERMODE_SILENT:
        *drivertype = canDRIVER_SILENT;
        break;
    case DRIVERMODE_SELFRECEPTION: 
        *drivertype = canDRIVER_SELFRECEPTION;
        break;
    case DRIVERMODE_OFF: 
        *drivertype = canDRIVER_OFF;
        break;
    }
    return canOK;

 nomem_error:
    close(fd);
    return canERR_NOMEM;
}

canStatus lpcAccept(HandleData *hData,
                    const long envelope,
                    const unsigned int flag)
{
    cmdSetHwFilterReq cmd;
    int ret;
    const long stdMask = (1 << 11) - 1;
    const long extMask = (1 << 29) - 1;

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo  = CMD_SET_SW_FILTER_REQ;
    cmd.transId = 0;
    cmd.channel = hData -> channelNr;
    cmd.flags = 0;

    switch (flag) {
    case canFILTER_SET_CODE_STD:
        cmd.acceptanceCode = (envelope & stdMask) << 0*(32-11);
        cmd.flags = FILTER_CODE_IS_VALID;
        break;
    case canFILTER_SET_MASK_STD:
        cmd.acceptanceMask = (~envelope & stdMask) << 0*(32-11);
        cmd.flags = FILTER_MASK_IS_VALID;
        break; 
    case canFILTER_SET_CODE_EXT:
        cmd.acceptanceCode = (envelope & extMask) << 0*(32-29);  
        cmd.flags = FILTER_CODE_IS_VALID | FILTER_EXTENDED_ID;  
        break;
    case canFILTER_SET_MASK_EXT:
        cmd.acceptanceMask = (~envelope & extMask) << 0*(32-29);
        cmd.flags = FILTER_MASK_IS_VALID | FILTER_EXTENDED_ID;
        break;
    default:
        return canERR_PARAM;
    }
    //    cmd.flags |= FILTER_SJA1000_SINGLE;
    do {
        ret = ioctl(hData -> fd, LAPCAN_IOC_SENDCMD, &cmd);
    } while (ret != 0 && errno == EAGAIN);

    if (ret != 0) return canERR_INVHANDLE;
    return canOK;
}

canStatus lpcWriteInternal(HandleData * hData, long id, void *msgPtr,
                           unsigned int dlc, unsigned int flag)
{
    cmdTxExtMessage tmpMsg;
    cmdTxStdMessage *stdPtr = (cmdTxStdMessage *)(&tmpMsg);
    cmdTxExtMessage *extPtr = &tmpMsg;
    int i;
    int ret;
    unsigned char sendExtended;

    if (flag & canMSG_STD) sendExtended = 0;
    else if (flag & canMSG_EXT) sendExtended =1;
    else sendExtended = hData -> isExtended;
  
    if (sendExtended){
        if (id >= (1 << 29)) return canERR_PARAM;
        extPtr -> cmdAndLen = CMDLEN_TX_EXT_MESSAGE_CMDANDLEN(dlc);
        extPtr -> channel_and_dlc = ((hData -> channelNr) << 4) + dlc;
        extPtr -> flags = 0;
        if (flag & canMSG_ERROR_FRAME) extPtr -> flags |= MSGFLAG_ERROR_FRAME;
        if (flag & canMSG_RTR) extPtr -> flags |= MSGFLAG_REMOTE_FRAME;
        /* No particular transId */
        extPtr -> transId = 0;
        extPtr -> id = id;
        /* Copy messagedata */
        if (msgPtr != NULL) {
            for (i=0; i < dlc; i++) extPtr -> data[i] = ((unsigned char *) msgPtr)[i];
        }
    } else {
        if (id >= (1 << 11)) return canERR_PARAM;
        stdPtr -> cmdAndLen = CMDLEN_TX_STD_MESSAGE_CMDANDLEN(dlc);
        stdPtr -> channel_and_dlc = ((hData -> channelNr) << 4) + dlc;
        stdPtr -> flags = 0;
        if (flag & canMSG_ERROR_FRAME) stdPtr -> flags |= MSGFLAG_ERROR_FRAME;
        if (flag & canMSG_RTR) stdPtr -> flags |= MSGFLAG_REMOTE_FRAME;
        /* No particular transId */
        stdPtr -> transId = 0;
        stdPtr -> id = id;
        /* Copy messagedata */
        if (msgPtr != NULL) {
            for (i=0; i < dlc; i++) stdPtr -> data[i] = ((unsigned char *)msgPtr)[i];
        }
    }
    ret = ioctl(hData -> fd, LAPCAN_IOC_SENDCMD, &tmpMsg);
    if (ret == 0) return canOK;
    else if (errno == EAGAIN) return canERR_TXBUFOFL;
    else if (errno == EBADMSG) return canERR_PARAM;
    /* Interrupted by signal */
    else if (errno == EINTR) return canERR_INTERRUPTED;
    else return canERR_PARAM;  
}

canStatus lpcWrite (HandleData *hData, long id, void *msgPtr,
                    unsigned int dlc, unsigned int flag)
{
  
    /* Set non blocking fileop */
    if (hData -> writeIsBlock) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_WRITE_BLOCK, 0);
    hData -> writeIsBlock = 0;
    return lpcWriteInternal(hData, id, msgPtr, dlc, flag);
}

canStatus lpcWriteWait (HandleData *hData, long id, void *msgPtr,
                        unsigned int dlc, unsigned int flag, long timeout)
{
    if (timeout == 0) return lpcWrite(hData, id, msgPtr, dlc, flag);
    /* Set non blocking fileop */
    if (!hData -> writeIsBlock) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_WRITE_BLOCK, 1);
    /* Set timeout in milliseconds */
    if (hData -> writeTimeout != timeout) 
        ioctl(hData -> fd, LAPCAN_IOC_SET_WRITE_TIMEOUT, timeout);

    return lpcWriteInternal(hData, id, msgPtr, dlc, flag);
}

canStatus lpcWriteSync (HandleData *hData, unsigned long timeout)
{
    int ret;
    ret = ioctl(hData -> fd, LAPCAN_IOC_WAIT_EMPTY, &timeout);
    if (ret == 0) return canOK;
    switch(errno){
    case EAGAIN:
        return canERR_TIMEOUT;
    case EINTR:
        return canERR_INTERRUPTED;
    default:
        return canERR_NOTFOUND;
    }
}

canStatus lpcReadTimer (HandleData *hData, unsigned long *time) {
    int fd;
    int ret;
    cmdReadClockReq cmd;
    cmdReadClockResp reply;
    fd = open(hData -> deviceName, O_RDONLY);

    /* Block everything but CMD_READ_CLOCK_RESP */
    ret = addFilter(fd, CMD_READ_CLOCK_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);

    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo = CMD_READ_CLOCK_REQ;
    cmd.transId = globalTransId++;
    ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);
    if (ret != 0) goto ioctl_error;
    while(1) {
        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) goto ioctl_error;
        if (reply.transId == cmd.transId){
            close(fd);
            if (time != NULL) *time = reply.currentTime / LAPCAN_TICKS_PER_MS; /* FIX 8 */
            return canOK;
        }
    }
 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
    return canERR_INVHANDLE;
}
canStatus lpcReadErrorCounters (HandleData *hData, unsigned int* txErr, 
                                unsigned int* rxErr, unsigned int* ovErr) 
{
    int fd;
    int ret;
    LapcanStat statBuffer;
    cmdGetChipState2Req cmd;
    cmdGetChipState2Resp reply;
    if (txErr != NULL || rxErr != NULL){
        fd = open(hData -> deviceName, O_RDONLY);
        /* Block everything but CMD_GET_CHIP_STATE2_RESP */
        ret = addFilter(fd, CMD_GET_CHIP_STATE2_RESP, -1, 0, 0, 0, 0, 1);
        if (ret != 0) goto nomem_error;
        ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
        if (ret != 0) goto nomem_error;
        ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);
    
        cmd.cmdLen = CMDLEN(cmd);
        cmd.cmdNo = CMD_GET_CHIP_STATE2_REQ;
        cmd.transId = globalTransId++;
        cmd.channel = hData -> channelNr; 
        ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);
        if (ret != 0) goto ioctl_error;

        ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
        if (ret != 0) goto ioctl_error;
        close(fd);
        if (txErr != NULL) *txErr = reply.txErrorCounter;
        if (rxErr != NULL) *rxErr = reply.rxErrorCounter;
    }
  
    if (ovErr != NULL){
        /* The driver will copy statBuffer.statSize bytes to us */
        statBuffer.statSize = sizeof(statBuffer);
        ret = ioctl(hData -> fd, LAPCAN_IOC_GET_STAT, &statBuffer);
        if (ret != 0) return canERR_INVHANDLE;
        *ovErr = statBuffer.overruns;
    }

    return canOK;
 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
    return canERR_INVHANDLE;
}
canStatus lpcReadStatus (HandleData *hData, unsigned long *flags) 
{
    int fd;
    int ret;
    LapcanStat statBuffer;
    cmdGetChipState2Req cmd;
    cmdGetChipState2Resp reply;
    if (flags == NULL) return canOK;

    fd = open(hData -> deviceName, O_RDONLY);
    /* Block everything but CMD_GET_CHIP_STATE2_RESP */
    ret = addFilter(fd, CMD_GET_CHIP_STATE2_RESP, -1, 0, 0, 0, 0, 1);
    if (ret != 0) goto nomem_error;
    ret = addFilter(fd, 0, 0, 0, 0, 0, 0, 0);
    if (ret != 0) goto nomem_error;
    ioctl(fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);
  
    cmd.cmdLen = CMDLEN(cmd);
    cmd.cmdNo = CMD_GET_CHIP_STATE2_REQ;
    cmd.transId = globalTransId++;
    cmd.channel = hData -> channelNr; 
    ret = ioctl(fd,LAPCAN_IOC_SENDCMD, &cmd);
    if (ret != 0) goto ioctl_error;
  
    ret = ioctl(fd, LAPCAN_IOC_RECVCMD, &reply);
    if (ret != 0) goto ioctl_error;
    close(fd);
    *flags = 0;
    if (reply.busStatus & BUSSTAT_BUSOFF) *flags |= canSTAT_BUS_OFF;
    if (reply.busStatus & BUSSTAT_ERROR_PASSIVE) *flags |= canSTAT_ERROR_PASSIVE;
    if (reply.busStatus &  BUSSTAT_ERROR_WARNING)  *flags |= canSTAT_ERROR_WARNING;
    if (reply.busStatus & BUSSTAT_ERROR_ACTIVE)  *flags |= canSTAT_ERROR_ACTIVE;
    if (reply.txErrorCounter != 0) *flags |= canSTAT_TXERR;
    if (reply.rxErrorCounter != 0) *flags |= canSTAT_RXERR;

    /* The driver will copy statBuffer.statSize bytes to us */
    statBuffer.statSize = sizeof(statBuffer);
    ret = ioctl(hData -> fd, LAPCAN_IOC_GET_STAT, &statBuffer);
    if (ret != 0) return canERR_INVHANDLE;
    if (statBuffer.overruns > 0) *flags |= canSTAT_SW_OVERRUN | canSTAT_HW_OVERRUN;
    if (statBuffer.rcvQL > 0) *flags |= canSTAT_RX_PENDING;
    if (statBuffer.sendQL > 0) *flags |= canSTAT_TX_PENDING;

    return canOK;
 nomem_error:
    close(fd);
    return canERR_NOMEM;
 ioctl_error:
    close(fd);
    return canERR_INVHANDLE;
}

canStatus lpcGetChannelData (int channel, int item, void *buffer, size_t bufsize)
{
    switch (item){

    default:
        return canERR_PARAM;
    }
}

canStatus lpcIoCtl(HandleData *hData, unsigned int func, 
                   void * buf, unsigned int buflen)
{
    int ret;
    LapcanStat statBuffer;
    cmdFlushQueue flushCmd;
    statBuffer.statSize = sizeof(statBuffer);
    switch(func) {
    case canIOCTL_GET_RX_BUFFER_LEVEL:
        // buf points at a DWORD which receives the current RX queue level. 
        ret = ioctl(hData -> fd, LAPCAN_IOC_GET_STAT, &statBuffer);
        if (buf == NULL) return canERR_PARAM;
        *(unsigned long*) (buf) = statBuffer.rcvQL;
        break;
    case canIOCTL_GET_TX_BUFFER_LEVEL:
        // buf points at a DWORD which receives the current TX queue level. 
        ret = ioctl(hData -> fd, LAPCAN_IOC_GET_STAT, &statBuffer);
        if (buf == NULL) return canERR_PARAM;
        *(unsigned long*) (buf) = statBuffer.sendQL;
        //printf("LAPfunc qlen= %d QL= %d (0x%x)\n", *(unsigned long*) (buf), statBuffer.sendQL, statBuffer.sendQL);
        break;
    case canIOCTL_FLUSH_RX_BUFFER:
        // Discard the current contents of the RX queue. 
        ioctl(hData -> fd, LAPCAN_IOC_FLUSH_RCVBUFFER, 0);
        break;
    case canIOCTL_FLUSH_TX_BUFFER:
        //  Discard the current contents of the TX queue. 
        flushCmd.cmdLen  = CMDLEN(flushCmd);
        flushCmd.cmdNo   = CMD_FLUSH_QUEUE;
        flushCmd.transId = globalTransId++;
        flushCmd.channel = hData -> channelNr; 
        flushCmd.flags   = FLUSH_TX_QUEUE;
        //    do{
        //  ret = ioctl(hData -> fd, LAPCAN_IOC_SENDCMD, &flushCmd);
        //} while (ret != 0 && errno == EAGAIN);
        //// ??? This is not working right!!!
        ret = ioctl(hData -> fd, LAPCAN_IOC_FLUSH_SENDBUFFER, 0);
        //printf("ioctl(hData -> fd, LAPCAN_IOC_FLUSH_SENDBUFFER, 0) ret= %d\n", ret);
        break;
    default:
        return canERR_PARAM;
    }
    return canOK;
}

HWOps lpcHWOps = {
    /* VCan Functions */
    setNotify: lpcSetNotify,
    openChannel: lpcOpenChannel,
    busOn: lpcBusOn,
    busOff: lpcBusOff,
    setBusParams: lpcSetBusParams,
    getBusParams: lpcGetBusParams,
    read: lpcRead,
    readWait: lpcReadWait,
    setBusOutputControl: lpcSetBusOutputControl,
    getBusOutputControl: lpcGetBusOutputControl,
    accept: lpcAccept,
    write: lpcWrite,
    writeWait: lpcWriteWait,
    writeSync: lpcWriteSync,
    readTimer: lpcReadTimer,
    readErrorCounters: lpcReadErrorCounters,
    readStatus: lpcReadStatus,
    getChannelData: lpcGetChannelData,
    ioCtl: lpcIoCtl
};













