/*  Kvaser Linux Canlib */
/*  Copyright (C) 2003 Kvaser AB - support@kvaser.com - www.kvaser.com  */

/*
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <canlib.h>

#include <lapcan_data.h>
#include <lapcan_ioctl.h>

#include "canlib_data.h"
#include "canlib_util.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <lapcmds.h>
#include <stdio.h>
#include <errno.h>
#include <signal.h>
#include <pthread.h>
#include <string.h>
#include <sys/stat.h>

#include "VCanFunctions.h"
#include "LapcanFunctions.h"

static HandleList *handleList;

static const char *errorStrings [] = {

    [-canOK]                    = "No error",
    [-canERR_PARAM]             = "Error in parameter",
    [-canERR_NOMSG]             = "No messages available",
    [-canERR_NOTFOUND]          = "Specified device not found",
    [-canERR_NOMEM]             = "Out of memory",
    [-canERR_NOCHANNELS]        = "No channels avaliable",
    [-canERR_INTERRUPTED]       = "Interrupted by signal",
    [-canERR_TIMEOUT]           = "Timeout occurred",
    [-canERR_NOTINITIALIZED]    = "Library not initialized",
    [-canERR_NOHANDLES]         = "No more handles",
    [-canERR_INVHANDLE]         = "Handle is invalid",
    [-canERR_DRIVER]            = "CAN driver type not supported",
    [-canERR_TXBUFOFL]          = "Transmit buffer overflow",
    [-canERR_HARDWARE]          = "A hardware error was detected",
    [-canERR_DRIVERLOAD]        = "Can not load or open the device driver",
    [-canERR_NOCARD]            = "Card not found"
};

/********************************************************/
/* Find out channel specific data                       */
/********************************************************/
canStatus getDevParams (int channel, char devName[], int *devChannel, HWOps **hwOps) {

    int chanCounter = 0;
    int devCounter = 0;
    int found = 0;
    struct stat stbuf;

    for (devCounter = 0; !found; (devCounter++, chanCounter++)){
        snprintf(devName, DEVICE_NAME_LEN, "/dev/pcican%d", devCounter); 
        if (stat(devName, &stbuf) == -1) break;
        if (chanCounter == channel) {
            *devChannel = 0;
            *hwOps = &vCanHWOps;
            found = 1;
        }
    }

    /* qqq If the VCanOsIf/LapcanOsIf driver is used, one device file per
           channel is created so the code for second channel below has to
           be removed
    */ 
    for (devCounter = 0; !found; (devCounter++, chanCounter++)){
        snprintf(devName, DEVICE_NAME_LEN, "/dev/lapcan%d", devCounter);    
        if (stat(devName, &stbuf) == -1) break;
        /* First Channel */
        if (chanCounter == channel) {
            *devChannel = 0;
            *hwOps = &lpcHWOps;
            found = 1;
        }
        /* Second Channel */
        chanCounter++;
        if (chanCounter == channel) {
            *devChannel = 1;
            *hwOps = &lpcHWOps;
            found = 1;
        }
    }
  
    if (found) {
        return canOK;
    } else {
        devName[0]  = 0;
        *devChannel = -1;
        *hwOps      = NULL;
        return canERR_NOTFOUND;
    }

}

inline static int hndCmp (const void * hData1, const void * hData2) 
{
    return ((HandleData*)(hData1)) -> handle == 
        ((HandleData*)(hData2)) -> handle;
}
inline static HandleData * findHandle (HandleList **handleList, CanHandle hnd)
{
    HandleData dummyHandleData;
    dummyHandleData.handle = hnd; 
    return listFind(handleList, &dummyHandleData, &hndCmp);
}

inline static HandleData * removeHandle (HandleList **handleList, CanHandle hnd)
{
    HandleData dummyHandleData;
    dummyHandleData.handle = hnd; 
    return listRemove(handleList, &dummyHandleData, &hndCmp);
}


/*
 * API FUNCTIONS
 */

CanHandle canOpenChannel (int channel, int flags)
{
    canStatus status;
    static CanHandle hnd = 0;
    HandleData *hData;

    hData = (HandleData*) malloc (sizeof(HandleData));
    memset(hData, 0, sizeof(HandleData));

    hData->readIsBlock      = 1;
    hData->writeIsBlock     = 1;
    hData->handle           = hnd;
    hData->isExtended       = flags & canWANT_EXTENDED;
    hData->wantExclusive    = flags & canWANT_EXCLUSIVE;

    status = getDevParams(channel, 
                          hData->deviceName,    
                          &hData->channelNr,    
                          &hData->hwOps);    
    if (status < 0) {
        free(hData);
        return status;
    }

    status = hData->hwOps->openChannel(hData);
    if (status < 0) {
        free(hData);
        return status;
    }
  
    listInsertFirst(&handleList, hData);
    return hnd++;
}


int canClose (const CanHandle hnd)
{
    HandleData *hData;
    hData = removeHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    canSetNotify(hnd, NULL, 0, NULL);  
    if (close(hData -> fd) != 0) return canERR_INVHANDLE;
    free(hData);
    return canOK;
}


canStatus canBusOn (const CanHandle hnd)
{
    HandleData *hData;
    hData = findHandle (&handleList, hnd);
    if (hData == NULL)  return canERR_INVHANDLE;
    return  (hData -> hwOps -> busOn(hData));
}


canStatus canBusOff (const CanHandle hnd)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> busOff(hData);
}


canStatus canSetBusParams (const CanHandle hnd, long freq, unsigned int tseg1,
                           unsigned int tseg2, unsigned int sjw, unsigned int noSamp,
                           unsigned int syncmode)
{
    canStatus ret; 
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    if (freq < 0) {
      ret = canTranslateBaud(&freq, &tseg1, &tseg2, &sjw, &noSamp, &syncmode);
      if (ret != canOK) return ret;
    }
    return hData->hwOps->setBusParams(hData, freq, tseg1, tseg2, sjw, noSamp, syncmode);
}


canStatus canGetBusParams(const CanHandle hnd, long * freq, unsigned int * tseg1,
                          unsigned int * tseg2, unsigned int * sjw,
                          unsigned int * noSamp, unsigned int * syncmode)
{
    HandleData *hData;
  
    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData->hwOps->getBusParams(hData, freq, tseg1, tseg2, sjw, noSamp, syncmode); 
}


canStatus canSetBusOutputControl (const CanHandle hnd, const unsigned int drivertype)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    if (drivertype != canDRIVER_NORMAL && drivertype != canDRIVER_OFF &&
        drivertype != canDRIVER_SILENT && drivertype != canDRIVER_SELFRECEPTION) {
        return canERR_PARAM;
    }
    return hData -> hwOps -> setBusOutputControl(hData, drivertype);
}


canStatus canGetBusOutputControl(const CanHandle hnd, unsigned int * drivertype)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> getBusOutputControl(hData, drivertype);
}


canStatus canAccept(const CanHandle hnd,
                    const long envelope,
                    const unsigned int flag)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> accept(hData, envelope, flag);
}


canStatus canReadStatus(const CanHandle hnd,
                        unsigned long * flags)
{
    HandleData *hData;
    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
  
    return hData -> hwOps -> readStatus(hData, flags);
}


canStatus canReadErrorCounters(CanHandle hnd,
                               unsigned int * txErr,
                               unsigned int * rxErr,
                               unsigned int * ovErr)
{
    HandleData *hData;
    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
  
    return hData -> hwOps -> readErrorCounters(hData, txErr, rxErr, ovErr);
}


canStatus canWrite (CanHandle hnd, long id, void *msgPtr,
                    unsigned int dlc, unsigned int flag)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> write(hData, id, msgPtr, dlc, flag);
}


canStatus canWriteWait (CanHandle hnd, long id, void *msgPtr,
                        unsigned int dlc, unsigned int flag, long timeout)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> writeWait(hData, id, msgPtr, 
                                               dlc, flag, timeout);
}


canStatus canRead (CanHandle hnd, long * id, void * msgPtr, unsigned int * dlc,
                   unsigned int * flag, unsigned long * time)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> read(hData, id, msgPtr, dlc, flag, time);
}


canStatus canReadWait (CanHandle hnd, long * id, void * msgPtr, unsigned int * dlc,
                       unsigned int * flag, unsigned long * time, long timeout)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    return hData -> hwOps -> 
        readWait(hData, id, msgPtr, dlc, flag, time, timeout);
}


canStatus canWriteSync(CanHandle hnd, unsigned long timeout)
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;

    return hData -> hwOps -> writeSync(hData, timeout);
}


canStatus canIoCtl(CanHandle hnd, unsigned int func, void * buf, unsigned int buflen)
{
    HandleData *hData;
    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;

    return hData -> hwOps -> ioCtl(hData, func, buf, buflen);
}


canStatus canReadTimer(CanHandle hnd, unsigned long *time)
{
    HandleData *hData;
    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
  
    return hData -> hwOps -> readTimer(hData, time);
}


canStatus canTranslateBaud(long * const freq,
                           unsigned int * const tseg1,
                           unsigned int * const tseg2,
                           unsigned int * const sjw,
                           unsigned int * const nosamp,
                           unsigned int * const syncMode)
{
    switch (*freq) {
    case BAUD_1M:
        *freq     = 1000000L;
        *tseg1    = 4;
        *tseg2    = 3;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_500K:
        *freq     = 500000L;
        *tseg1    = 4;
        *tseg2    = 3;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_250K:
        *freq     = 250000L;
        *tseg1    = 4;
        *tseg2    = 3;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_125K:
        *freq     = 125000L;
        *tseg1    = 10;
        *tseg2    = 5;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_100K:
        *freq     = 100000L;
        *tseg1    = 10;
        *tseg2    = 5;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_62K:
        *freq     = 62500L;
        *tseg1    = 10;
        *tseg2    = 5;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    case BAUD_50K:
        *freq     = 50000L;
        *tseg1    = 10;
        *tseg2    = 5;
        *sjw      = 1;
        *nosamp   = 1;
        *syncMode = 0;
        break;
    default:
        return canERR_PARAM;
    }
    return canOK;
}


canStatus canGetErrorText(canStatus err,
                          char * buf,
                          unsigned int bufsiz)
{
    signed char code;

    code = (signed char)(err & 0xFF);

    if (!buf || bufsiz == 0) return canOK;
    if ((code <= 0) && (-code < sizeof(errorStrings)/sizeof(char*))) {
        if (errorStrings [-code] == NULL){
            snprintf(buf, bufsiz, "Unknown error (%d)", (int)code);
        } else {
            strncpy(buf, errorStrings[-code], bufsiz);  
        }
    } else {
        strncpy(buf, "This is not an error code", bufsiz);
    }
    buf[bufsiz-1] = '\0';  
    return canOK;
}

    
unsigned short canGetVersion (void)
{
    return (CANLIB_MAJOR_VERSION<<8) + CANLIB_MINOR_VERSION;
}


canStatus canGetNumberOfChannels (int * channelCount)
{
    /*
     * qqq This function has to be modified if we have other cards than LAPcan.
     * For now, we just count the number of /dev/lapcan* files and multiply
     * with 2 
    */ 

    int tmpCount = 0;
    int cardNr;
    char filename[DEVICE_NAME_LEN+1];
    for(cardNr = 0; cardNr <= 255; cardNr++) { /* There are 256 minor inode numbers */ 
        snprintf(filename,  DEVICE_NAME_LEN, "/dev/lapcan%d", cardNr);
        if (!access(filename, F_OK)) tmpCount++;  /* Check for existance */
    }
    *channelCount = tmpCount * 2;
    return canOK;
}


canStatus canGetChannelData (int channel, int item, void *buffer, size_t bufsize)
{
    canStatus status;
    HandleData hData;
    status = getDevParams(channel, 
                          hData.deviceName,    
                          &hData.channelNr,    
                          &hData.hwOps);    
    if (status < 0) return status;
      
    switch(item){
    default:

        return hData.hwOps->getChannelData(channel, item, buffer, bufsize);
    }
}


canStatus canSetNotify (const CanHandle hnd, void (*callback) (canNotifyData *), 
                        unsigned int notifyFlags, void * tag)
     /* 
      * Notification is done by filtering out interresting messages and
      * doing a blocked read from a thread
     */
{
    HandleData *hData;

    hData = findHandle (&handleList, hnd);
    if (hData == NULL) return canERR_INVHANDLE;
    if (notifyFlags == 0 || callback == NULL) {
        /* We want to shut off notification, close file and clear callback */ 
        pthread_cancel(hData -> notifyThread);
        /* Wait for thread to finish */
        pthread_join(hData -> notifyThread, NULL);
    
        if (hData -> notifyFd != 0) close(hData -> notifyFd);
        hData -> notifyFd = 0;
        callback = NULL;
        return canOK;
    }
    hData -> notifyData.tag = tag;
    return hData -> hwOps -> setNotify(hData, callback, notifyFlags);
}


void canInitializeLibrary()
{
    return;
}













