/*  LAPcan driver kernel module for Linux */
/*  Copyright (C) 2003 Kvaser AB - support@kvaser.com - www.kvaser.com  */

/* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or 
* (at your option) any later version.
* 
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* 
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
* 
*/

#include <linux/config.h>

#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/slab.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/ioport.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <asm/uaccess.h>

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ds.h>
#include <pcmcia/bus_ops.h>

//#include <sysdep.h> /* Some compatibility with older kernels qqq*/

#include <lapcan_main_cs.h>
#include <lapcan_ioctl.h>
#include <tl16564.h>

#define put_user_ret(x,ptr,ret) ({ if (put_user(x,ptr)) return ret; })
#define get_user_ret(x,ptr,ret) ({ if (get_user(x,ptr)) return ret; })
#define copy_to_user_ret(to,from,n,retval) ({ if (copy_to_user(to,from,n)) return retval; })
#define copy_from_user_ret(to,from,n,retval) ({ if (copy_from_user(to,from,n)) return retval; })

/* Module versioning */

#define EXPORT_SYMTAB
#include <linux/module.h>


#include <linux/autoconf.h> /* retrieve the CONFIG_* macros */
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#   define MODVERSIONS
#endif  

#ifdef MODVERSIONS
#   include <linux/modversions.h>
#endif


/*
All the PCMCIA modules use PCMCIA_DEBUG to control debugging.  If
you do not define PCMCIA_DEBUG at all, all the debug code will be
left out.  If you compile with PCMCIA_DEBUG=0, the debug code will
be present but disabled -- but it can then be enabled for specific
modules at load time with a 'pc_debug=#' option to insmod.
*/


#ifdef PCMCIA_DEBUG
static int pc_debug = PCMCIA_DEBUG;
MODULE_PARM(pc_debug, "i");
//#define DEBUG(n, args...) if (pc_debug>(n)) printk( KERN_DEBUG args)
#   define DEBUG(n, args...) if (pc_debug>=(n)) printk("<" #n ">" args)
static char *version =
"LAPCan version 1.0.1 $Revision: 1.8 $";
#else
#   define DEBUG(n, args...) 

#endif

static char *device_name = DEVICE_NAME_STRING;


/*======================================================================*/

/* Parameters that can be set with 'insmod' */

/* Release IO ports after configuration? */
static int free_ports = 0;

/* The old way: bit map of interrupts to choose from */
/* This means pick from 15, 14, 12, 11, 10, 9, 7, 5, 4, and 3 */ 
static u_int irq_mask = 0xdeb8;
/* Newer, simpler way of listing specific interrupts */
static int irq_list[4] = { -1 };

MODULE_PARM(free_ports, "i");
MODULE_PARM(irq_mask, "i");
MODULE_PARM(irq_list, "1-4i");

/*======================================================================*/

/*
The event() function is this driver's Card Services event handler.
It will be called by Card Services when an appropriate card status
event is received.  The config() and release() entry points are
used to configure or release a socket, in response to card
insertion and ejection events.  They are invoked from the lapcan
event handler. 
*/

static void lapcanConfig(dev_link_t *link);
static void lapcanRelease(u_long arg);
static int  lapcanEvent(event_t event, int priority,
                        event_callback_args_t *args);

                        /*
                        The attach() and detach() entry points are used to create and destroy
                        "instances" of the driver, where each instance represents everything
                        needed to manage one actual PCMCIA card.
*/

static dev_link_t *lapcanAttach(void);
static void lapcanDetach(dev_link_t *);
static void lapcanTransmit(LapcanContext *ctxPtr, lpcCmd *msg);
int lapcanOpen (struct inode *inode, struct file *filp);
int lapcanClose (struct inode *inode, struct file *filp);
int lapcanIOCtl (struct inode *inode, struct file *filp, 
                 unsigned int cmd, unsigned long arg);

static void lapcanReceive (LocalInfo *local);
static void lapcanSend (void *void_localPtr);
static int waitTransmit (LapcanContext *ctxPtr);
static int lapcanInit(LapcanContext *ctxPtr);


/*
The devInfo variable is the "key" that is used to match up this
device driver with appropriate cards, through the card configuration
database.
*/

static dev_info_t devInfo = DEVICE_NAME_STRING;



/* List of all lapcan cards(devices) */
static dev_link_t *devList = NULL;

struct file_operations lapcanFOps = {
llseek:    NULL,        
read:      NULL,        
write:     NULL,        
readdir:   NULL,        
poll:      NULL,        
ioctl:     lapcanIOCtl, 
mmap:      NULL,        
open:      lapcanOpen,  
flush:     NULL,        
release:   lapcanClose, 
fsync:     NULL        
           /* Fills rest with NULL entries */
};

unsigned char chanMap [CMD__HIGH + 1] = \
{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 2, 3, 3, 5, \
3, 3, 0, 0, 3, 0, 3, 3, 3, 3, \
3, 3, 3, 3, 0, 0, 0, 0, 0, 0, \
0, 0, 3, 3, 0, 0, 0, 3, 0, 3, \
3, 0, 3, 3, 3, 0, 0, 0, 3, 3, \
3, 3, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 3, 3, 2, \
3, 3, 0, 0, 0, 3, 0, 0, 0, 3, \
3, 3, 3};

unsigned char flagMap [CMD__HIGH + 1] = \
{2, 2, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 4, 6, 0, \
0, 0, 0, 0, 6, 0, 0, 0, 0, 0, \
0, 6, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \

0, 0, 0, 6, 5, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, \
4, 0, 0};


/*======================================================================*/
/* getCmdNr                                                             */
/*======================================================================*/

inline int getCmdNr (lpcCmd *cmd)
{
    unsigned char *b;  
    
    b = (unsigned char*) cmd;
    if (b [0] & 0x80) { /* Infrequent command:bit 7 is set */
        return b [1] & 0x7f;
    } 
    else {
        return b [0] >> 5;
    }
}

/*======================================================================*/
/* Find out length of lapcan command                                    */
/*======================================================================*/

inline int getCmdLen (lpcCmd *cmd)
{  
    unsigned char *b;
    
    b = (unsigned char*) cmd;
    if (b [0] & 0x80) { /* Infrequent command:bit 7 is set */
        return 1 + (b [0] & 0x7f); /* Length is lower 7 bits */
    } 
    else {
        return 1 + (b [0] & 0x1f); /* Length is lower 5 bits */
    }
}

/*======================================================================*/
/*    getCmdChannel                                                     */
/*======================================================================*/

inline int getCmdChan (lpcCmd *cmd)
{
    unsigned char *b;
    b = (unsigned char*) cmd;
    if (b [0] & 0x80) { /* Infrequent command:bit 7 is set */
        return b [chanMap [getCmdNr(cmd)] ];
    } 
    else {
        return b [1] >> 4; /* channel is upper 4 bits */
    }
}

inline unsigned char getFlags (lpcCmd *cmd)
{
    /* flagMap[getCmdNr(cmd)] is the position of the flags in the command */
    return ((unsigned char*) cmd) [flagMap [getCmdNr(cmd)] ];
}

/*======================================================================*/
/* setFlags                                                             */
/*======================================================================*/

inline void setFlags (lpcCmd *cmd, unsigned char flags)
{
    ((unsigned char*) cmd) [flagMap [getCmdNr(cmd)] ] = flags;
    return;
}

/*======================================================================*/
/* Copy a lapcan command                                                */
/*======================================================================*/

inline void copyLapcanCmd (lpcCmd *cmd_to, lpcCmd *cmd_from)
{
    int cmdLen;
    unsigned char *from, *to;
    
    from = (unsigned char*) cmd_from;
    to = (unsigned char*) cmd_to;
    
    for (cmdLen = getCmdLen(cmd_from); cmdLen > 0; cmdLen--) {
        *to++ = *from++;
    }
    return;
}

/*======================================================================*/
/* Filter for lapcan commands                                           */
/*======================================================================*/

int addFilter (LapcanOpenFileNode *openFileNodePtr, FilterData *filterDataPtr)
//int addFilter (int *openFileNodePtr, int *filterDataPtr)
{
    FilterNode **filterNodePtrPtr;
    FilterNode *newFilterNodePtr;
    newFilterNodePtr = kmalloc(sizeof(FilterNode), GFP_KERNEL);
    if (!newFilterNodePtr) return -ENOMEM;
    filterNodePtrPtr = &(openFileNodePtr -> msgFilter);
    /* Find end of filterlist */
    while (*filterNodePtrPtr != NULL) filterNodePtrPtr = &((*filterNodePtrPtr) -> next);
    /* Insert into filterlist */
    newFilterNodePtr -> cmdNr = filterDataPtr -> cmdNr;
    newFilterNodePtr -> cmdNrMask = filterDataPtr -> cmdNrMask;
    newFilterNodePtr -> chanNr = filterDataPtr -> chanNr;
    newFilterNodePtr -> chanNrMask = filterDataPtr -> chanNrMask;
    newFilterNodePtr -> flags = filterDataPtr -> flags;
    newFilterNodePtr -> flagsMask = filterDataPtr -> flagsMask;
    newFilterNodePtr -> isPass = filterDataPtr -> isPass;  
    newFilterNodePtr -> next = NULL;
    *filterNodePtrPtr = newFilterNodePtr;  
    return 0;
}
/*======================================================================*/
/* Remove filters                                                       */
/*======================================================================*/

int clearFilters (LapcanOpenFileNode *openFileNodePtr)
{
    FilterNode **filterNodePtrPtr;
    FilterNode **tmpNodePtrPtr;
    filterNodePtrPtr = &(openFileNodePtr -> msgFilter);
    while (*filterNodePtrPtr != NULL) {
        tmpNodePtrPtr = filterNodePtrPtr;
        filterNodePtrPtr = &((*filterNodePtrPtr) -> next);
        kfree(*tmpNodePtrPtr);
        *tmpNodePtrPtr = NULL;
    }
    return 0;
}

/*======================================================================*/
/* First, check if the pass filter matches. If so, return pass.         */
/* Then check if block filter matches and return block in that case.    */
/*    Otherwise, return pass.                                           */
/*    Condition: ((value XOR filter) AND filtermask) OR Condition       */
/*======================================================================*/

int checkFilters (LapcanOpenFileNode *openFileNodePtr, lpcCmd *cmdPtr)
{
    FilterNode *filterNodePtr;
    filterNodePtr = openFileNodePtr -> msgFilter;
    while (filterNodePtr != NULL) {
        
        if ( ! (((filterNodePtr -> cmdNr  ^ getCmdNr(cmdPtr))   & filterNodePtr -> cmdNrMask)  | \
            ((filterNodePtr -> chanNr ^ getCmdChan(cmdPtr)) & filterNodePtr -> chanNrMask) | \
            ((filterNodePtr -> flags  ^ getFlags(cmdPtr))   & filterNodePtr -> flagsMask))) {
            /* If this equals zero => check whether to pass or block */
            if (filterNodePtr -> isPass) return -1;
            else return 0;
        }
        filterNodePtr = filterNodePtr -> next;
    }
    /* No match => pass */
    
    return -1;  
}

/*======================================================================*/
/*  Calculate queue length                                              */
/*======================================================================*/

inline static int getQLen(int head, int tail, int size)
{
    return (head < tail ? head - tail + size : head - tail);
}

/*======================================================================*/
/*    lapcanGetStat - Return some statusinformation                     */
/*======================================================================*/

int lapcanGetStat (CanIfStat *userStatPtr, struct file *filp)
{
    unsigned int msgCount = 0;
    unsigned int tail;
    int channelNr;
    unsigned short size;
    unsigned char cmdNr;
    CanIfStat status;
    LapcanOpenFileNode *fileNodePtr;
    LapcanContext *ctxPtr;
    int * bufIndexPtr;
    lpcCmd * bufCmdPtr;
    unsigned int tmpQLen = 0;
    
    copy_from_user_ret(&status, userStatPtr, sizeof(status.statSize), -EFAULT);
    size = status.statSize;
    fileNodePtr = (LapcanOpenFileNode*) (filp -> private_data);
    ctxPtr = &fileNodePtr -> localPtr -> ctx;
    tail = fileNodePtr -> rcvBufTail;
    while (fileNodePtr -> rcvBufHead != tail) {
        cmdNr = getCmdNr(&fileNodePtr -> fileRcvBuffer[tail]);
        if (cmdNr == CMD_RX_STD_MESSAGE || cmdNr == CMD_RX_EXT_MESSAGE) msgCount++;
        tail++;
        if (tail >= FILE_RCV_BUF_SIZE) tail = 0;
    }
    status.rcvQL = msgCount;
    
    channelNr = fileNodePtr -> chanNr;
    if (channelNr == -1) return -EINVAL;
    //status.sendQL = getQLen(ctxPtr -> sentTXflagCount [channelNr], 
    //    ctxPtr -> recvTXflagCount [channelNr], TX_CHAN_BUF_SIZE/*1 << sizeof(int)*/);
    
    
    bufIndexPtr = &(ctxPtr -> txChanBufHead[channelNr]);
    bufCmdPtr = (lpcCmd*) &(ctxPtr -> txChanBuffer[channelNr][*bufIndexPtr]
        );
    //copyLapcanCmd(bufCmdPtr, &tmpCmd);
    //printk(" ******** Q ch %d head %d",  getCmdChan(bufCmdPtr), *bufIndexPtr);
    
//    status.sendQL = (getQLen(ctxPtr -> txChanBufHead[channelNr], ctxPtr -> txChanBufTail[channelNr], TX_CHAN_BUF_SIZE));
    tmpQLen = getQLen(ctxPtr -> sentTXflagCount [channelNr], 
         ctxPtr -> recvTXflagCount [channelNr], /*TX_CHAN_BUF_SIZE*/ 1 << sizeof(int));
    
    DEBUG(1, "******** LAPCAN TX count = %d\n", tmpQLen);
    status.sendQL = (getQLen(ctxPtr -> txChanBufHead[channelNr], ctxPtr -> txChanBufTail[channelNr], TX_CHAN_BUF_SIZE));
    DEBUG(1, "******** DRIVER TX count = %d\n", status.sendQL);
    status.sendQL = tmpQLen;
    
    status.overruns = fileNodePtr -> overruns [channelNr];
    copy_to_user_ret(userStatPtr, &status, sizeof(CanIfStat), -EFAULT);
    return 0;
}

/*======================================================================*/
/*    Open - File operation                                             */
/*======================================================================*/

int lapcanOpen(struct inode *inode, struct file *filp)
{
    
    LapcanOpenFileNode *openFileNodePtr;
    LocalInfo *localPtr;
    int minorNr;
    
    minorNr = MINOR(inode -> i_rdev);
    if (devList == NULL) return -ENODEV;
    localPtr = devList -> priv;
    
    /* Find the right minor inode number */
    while (localPtr != NULL && localPtr -> node1.minor != minorNr)
        localPtr = localPtr -> link.next -> priv;
    if (localPtr == NULL) return -ENODEV;
    
    /* The open-field holds the usage count for this lapcan*/
    localPtr -> link.open++;
    DEBUG(6,"Found inode %d, has %d open files\n",minorNr, localPtr -> link.open);
    
    /* Allocate memory */
    openFileNodePtr = kmalloc(sizeof(LapcanOpenFileNode), GFP_KERNEL);  
    if (!openFileNodePtr) return -ENOMEM;
    memset(openFileNodePtr, 0, sizeof(LapcanOpenFileNode));
    
    /* Init wait queue */
    init_waitqueue_head (&(openFileNodePtr -> rxWaitQ));
    
    
    /* Set read buffer position */ 
    openFileNodePtr -> rcvBufTail = 0;
    openFileNodePtr -> rcvBufHead = 0;
    openFileNodePtr -> filp = filp;
    openFileNodePtr -> localPtr = localPtr;
    openFileNodePtr -> writeIsBlock = 1;
    openFileNodePtr -> readIsBlock = 1;
    openFileNodePtr -> chanNr = -1;
    openFileNodePtr -> writeTimeout = -1;
    openFileNodePtr -> readTimeout = -1;
    
    /* Insert this node first in list of "opens" */
    openFileNodePtr -> next = localPtr -> ctx.openFileList;
    localPtr -> ctx.openFileList = openFileNodePtr;
    
    /* We want a pointer to the node as private_data */
    filp -> private_data = openFileNodePtr;
    /* DEBUG
    for(i=1; openFileNodePtr != NULL; i++){
    DEBUG(1, "OPEN FILE %d, filp = 0x%x\n", i, (int) openFileNodePtr -> filp);
    openFileNodePtr = openFileNodePtr -> next;
    }
    */
    
#if 0
    MOD_INC_USE_COUNT;
#endif
    return 0; 
}

/*======================================================================*/
/*   Release - File operation                                           */
/*======================================================================*/

int lapcanClose (struct inode *inode, struct file *filp) {
    LocalInfo *localPtr;
    LapcanOpenFileNode **openPtrPtr;
    LapcanOpenFileNode *fileNodePtr;
    LapcanContext *ctxPtr;
#if 0
    MOD_DEC_USE_COUNT;
#endif
    localPtr = ((LapcanOpenFileNode*) (filp -> private_data)) -> localPtr;
    ctxPtr = &localPtr -> ctx;
    /* Decrease usage count for this lapcan */
    localPtr -> link.open--;
    /* Find the open file node */
    openPtrPtr = &localPtr -> ctx.openFileList;
    for(;*openPtrPtr != NULL; openPtrPtr = &((*openPtrPtr) -> next)) {
        if ((*openPtrPtr) -> filp == filp) break;
    }
    /* We did not find anything? */
    if (*openPtrPtr == NULL) return -EBADF;
    /* openPtrPtr now points to the next-pointer that points to the correct node */
    fileNodePtr = *openPtrPtr; 
    /* Clear exclusive rights & open flags*/
    if ( fileNodePtr -> chanNr != -1 ) { 
        (ctxPtr -> channelOpenCount [fileNodePtr -> chanNr])--;
        /* Clear exclusive flag */
        ctxPtr -> exclusiveFlagPtr [fileNodePtr -> chanNr] = NULL;
        fileNodePtr -> chanNr = -1; 
    }
    
    /* Remove node */
    *openPtrPtr = (*openPtrPtr) -> next;
    //  DEBUG(1, "CLOSED %p\n", filp);
    /* Release memory allocated by filters */
    clearFilters(fileNodePtr);
    kfree(fileNodePtr);
    
    return 0;
}

/*======================================================================*/
/*   Poll - File operation                                              */
/*======================================================================*/
/*
unsigned int lapcanPoll (struct file *filp, poll_table *wait)
{
LapcanOpenFileNode *fileNodePtr;
LapcanContext *ctxPtr;

  fileNodePtr = filp -> private_data;
  localPtr = fileNodePtr -> localPtr;
  ctxPtr = &localPtr -> ctx;
  
    poll_wait(&fileNodePtr -> rxWaitQ, wait);
    poll_wait(&ctxPtr -> txChanWaitQ [chanNr];
    }
*/

/*======================================================================*/
/* IOCtl - File operation                                               */
/* This function is not reentrant with the same file descriptor!        */
/*======================================================================*/

int lapcanIOCtl (struct inode *inode, struct file *filp, 
                 unsigned int ioctl_cmd, unsigned long arg)
{
    LocalInfo *localPtr;
    LapcanOpenFileNode *fileNodePtr;
    lpcCmd *userCmdPtr;
    lpcCmd *bufCmdPtr;
    int *bufIndexPtr;
    lpcCmd tmpCmd;
    FilterData tmpFilterData;
    LapcanContext *ctxPtr;
    int chanNr;
    unsigned char cmdNr;
    unsigned char cmdLen;
    unsigned long timeLeft;
    unsigned long timeout;
    int i;
    
    fileNodePtr = filp -> private_data;
    localPtr = fileNodePtr -> localPtr;
    ctxPtr = &localPtr -> ctx;
    switch (ioctl_cmd) 
    {
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SENDCMD:
        userCmdPtr = (lpcCmd*) arg;
        /* We return -EFAULT directly if this fails */
        copy_from_user_ret(&tmpCmd, userCmdPtr, 2, -EFAULT);
        cmdLen = getCmdLen(&tmpCmd);
        copy_from_user_ret(&tmpCmd, userCmdPtr, cmdLen, -EFAULT);
        cmdNr = getCmdNr(&tmpCmd);
        switch (cmdNr) {
        case CMD_TX_STD_MESSAGE:
        case CMD_TX_EXT_MESSAGE:
            chanNr = getCmdChan(&tmpCmd);
            if (chanNr >= NR_CHANNELS) return -EBADMSG;
            
            /* If Buffer full and block is not set, return -EAGAIN */
            if (!fileNodePtr -> writeIsBlock && 
                getQLen(ctxPtr -> txChanBufHead[chanNr], ctxPtr -> txChanBufTail[chanNr],
                TX_CHAN_BUF_SIZE) >= TX_CHAN_BUF_SIZE - 1) { 
                return -EAGAIN;
            }
            /* Sleep when buffer is full, disable interrupts to avoid sleep race 
            * Variant : Put ourselves to sleep use locks ( read_lock() in im.q.) */
            while(1) {
                cli();
                if (getQLen(ctxPtr -> txChanBufHead [chanNr], ctxPtr -> txChanBufTail[chanNr],
                    TX_CHAN_BUF_SIZE) >= TX_CHAN_BUF_SIZE - 1)
                {
                    /* Do we want a timeout ?*/
                    if (fileNodePtr -> writeTimeout != -1){
                        if (interruptible_sleep_on_timeout(&ctxPtr->txChanWaitQ[chanNr], fileNodePtr->writeTimeout * HZ/1000) == 0) {
                            /* Sleep was interrupted by timer */
                            sti();
                            return -EAGAIN;
                        }
                    } else {
                        interruptible_sleep_on(&ctxPtr -> txChanWaitQ[chanNr]);
                    }
                    if (signal_pending(current)) {sti(); return -ERESTARTSYS;}
                    continue;
                } else {
                    /* Get a pointer to the right bufferspace */
                    /* ??? We need a lock here for SMP */
                    bufIndexPtr = &(ctxPtr -> txChanBufHead[chanNr]);
                    bufCmdPtr = (lpcCmd*) &(ctxPtr -> txChanBuffer[chanNr][*bufIndexPtr]);
                    copyLapcanCmd(bufCmdPtr, &tmpCmd);
                    DEBUG(8,"Q cmd %d @ %d head %d", getCmdNr(bufCmdPtr), 
                        getCmdChan(bufCmdPtr), *bufIndexPtr);
                    (*bufIndexPtr)++;
                    if (*bufIndexPtr >= TX_CHAN_BUF_SIZE) (*bufIndexPtr) = 0;
                    break;
                }
            }
            setFlags(bufCmdPtr, getFlags(bufCmdPtr) | MSGFLAG_TX);
            /* THIS REQUIRES MSGFLAG_TX TO BE SET!!! */
            ctxPtr -> sentTXflagCount [chanNr]++; 
            /* Add lapcanSend to the immediate task queue */
            if (ctxPtr -> fifoCount == 0){ 
                queue_task(&ctxPtr -> txTaskQ, &tq_immediate);
                mark_bh(IMMEDIATE_BH);
            }
            sti(); /* Late ??? */
            break;
            
        default: /* All other messages */
            
            /* If Buffer full and nonblock is set, return -EAGAIN */
            if (!fileNodePtr -> writeIsBlock && 
                getQLen(ctxPtr -> txCmdBufHead, ctxPtr -> txCmdBufTail,
                TX_CMD_BUF_SIZE) >= TX_CMD_BUF_SIZE - 1){ 
                return -EAGAIN;
            }
            /* Sleep when buffer is full, disable interrupts to avoid sleep race */
            while(1) {
                cli();
                if (getQLen(ctxPtr -> txCmdBufHead, ctxPtr -> txCmdBufTail,
                    TX_CMD_BUF_SIZE) >= TX_CMD_BUF_SIZE - 1)
                {
                    /* Do we want a timeout ?*/
                    if (fileNodePtr -> writeTimeout != 0){
                        if (interruptible_sleep_on_timeout(&ctxPtr -> txCmdWaitQ, 
                            fileNodePtr -> writeTimeout * HZ/1000)
                            == 0) {
                            /* Sleep was interrupted by timer */
                            sti();
                            return -EAGAIN;
                        }
                    } else {
                        interruptible_sleep_on(&ctxPtr -> txCmdWaitQ);
                    }
                    if (signal_pending(current)) {sti(); return -ERESTARTSYS;}
                    continue;
                } else {
                    /* Get a pointer to the right bufferspace */
                    /* Lock needed for SMP */
                    bufIndexPtr = &(ctxPtr -> txCmdBufHead);
                    bufCmdPtr = (lpcCmd*) &(ctxPtr -> txCmdBuffer [*bufIndexPtr]);
                    copyLapcanCmd(bufCmdPtr, &tmpCmd);
                    (*bufIndexPtr)++;
                    if (*bufIndexPtr >= TX_CMD_BUF_SIZE) (*bufIndexPtr) = 0;
                    sti();
                    break;
                }
            }
            /* --- */
            
            if (ctxPtr -> fifoCount ==0) {
                queue_task(&ctxPtr -> txTaskQ, &tq_immediate);
                mark_bh(IMMEDIATE_BH);
            }
            break;
        }
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_RECVCMD:
        /* If Buffer full and nonblock is set, return -EAGAIN */
        if (!fileNodePtr -> readIsBlock && 
            getQLen(fileNodePtr -> rcvBufHead, fileNodePtr -> rcvBufTail,
            FILE_RCV_BUF_SIZE) == 0) { 
            return -EAGAIN;
        }
        /* Sleep when buffer is full, disable interrupts to avoid sleep race */
        while(1) {
            cli();
            if (getQLen(fileNodePtr -> rcvBufHead, fileNodePtr -> rcvBufTail,
                FILE_RCV_BUF_SIZE) == 0) {
                /* Do we want a timeout, we add 1 to timeout so we don't sleep too short */
                
                if (fileNodePtr -> readTimeout != -1){
                    if (interruptible_sleep_on_timeout(&fileNodePtr -> rxWaitQ, 
                        1 + fileNodePtr -> readTimeout * HZ/1000)
                        == 0) {
                        /* Sleep was interrupted by timer */
                        sti();
                        return -EAGAIN;
                    }
                } else {
                    interruptible_sleep_on(&fileNodePtr -> rxWaitQ);
                }
                if (signal_pending(current)) {sti(); return -ERESTARTSYS;}
                continue;
            } else {
            /* Get a pointer to the right bufferspace 
                * ??? Lock needed for SMP */
                bufCmdPtr = &(fileNodePtr -> fileRcvBuffer [fileNodePtr -> rcvBufTail]);
                if (++(fileNodePtr -> rcvBufTail) >= FILE_RCV_BUF_SIZE) 
                    fileNodePtr -> rcvBufTail = 0;
                sti();
                break;
            }
        }
        DEBUG(8,"Read cmd %d filp = %p, nexttime = %d\n", 
            
            getCmdChan(bufCmdPtr), fileNodePtr -> filp, fileNodePtr -> rcvBufTail);
        userCmdPtr = (lpcCmd*) arg;
        copy_to_user_ret(userCmdPtr, bufCmdPtr, getCmdLen(bufCmdPtr), -EFAULT);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SET_EXCL:
        get_user_ret(chanNr, (int*) arg, -EFAULT);
        /* Return channel= -1 if channel is opened by someone else */
        spin_lock(&ctxPtr -> exclAndOpenLock [chanNr]);
        if (fileNodePtr -> chanNr == chanNr &&
            ctxPtr -> channelOpenCount [chanNr] == 1) {
            ctxPtr -> exclusiveFlagPtr [chanNr] = filp;
            spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);
        } else {
            spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);	
            chanNr = -1;
        }
        put_user_ret(chanNr, (int*) arg, -EFAULT);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_CLR_EXCL:
        get_user_ret(chanNr, (int*) arg, -EFAULT);
        /* Return channel=-1 if we don't own the channel */
        spin_lock(&ctxPtr -> exclAndOpenLock [chanNr]);
        if (ctxPtr -> exclusiveFlagPtr [chanNr] == NULL || 
            ctxPtr -> exclusiveFlagPtr [chanNr] == filp) {
            ctxPtr -> exclusiveFlagPtr [chanNr] = NULL;
            spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);
        } else { 
            DEBUG(1, "exclusiveFlagPtr [%d] == $%X\n", chanNr,
                ctxPtr -> exclusiveFlagPtr [chanNr] == NULL);
            spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);
            chanNr = -1;
        }
        put_user_ret(chanNr, (int*) arg, -EFAULT);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_OPEN_CHAN:
        get_user_ret(chanNr, (int*) arg, -EFAULT);
        spin_lock(&ctxPtr -> exclAndOpenLock [chanNr]);
        if (fileNodePtr -> chanNr == chanNr) {
            ; /* Already open. Do nothing */
        } else if (ctxPtr -> exclusiveFlagPtr [chanNr] == NULL) {
            /* Mark this channel as used */
            (ctxPtr -> channelOpenCount [chanNr])++;
            fileNodePtr -> chanNr = chanNr;
        } else {
            /* Someone has open this channel exclusively */
            chanNr = -1;
        }
        spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);
        put_user_ret(chanNr, (int*) arg, -EFAULT);	
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_CLOSE_CHAN:
        get_user_ret(chanNr, (int*) arg, -EFAULT);
        spin_lock(&ctxPtr -> exclAndOpenLock [chanNr]);
        if (fileNodePtr -> chanNr == chanNr ) { 
            (ctxPtr -> channelOpenCount [chanNr])--;
            fileNodePtr -> chanNr = -1;
            /* Clear exclusive flag */
            ctxPtr -> exclusiveFlagPtr [chanNr] = NULL; 
        }
        spin_unlock(&ctxPtr -> exclAndOpenLock [chanNr]);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_WAIT_EMPTY:
        get_user_ret(timeout, (unsigned long*) arg, -EFAULT);
        i = fileNodePtr -> chanNr;
        if (i == -1) return -EINVAL;
        /* Do we need to wait? */
        if (ctxPtr -> sentTXflagCount [i] > ctxPtr -> recvTXflagCount [i]){
            cli();
            /* Sleep for [timeout] milliseconds
            * i.s.o.t will re-enable interrupts before going to sleep */
            timeLeft = interruptible_sleep_on_timeout(&(ctxPtr -> flushQ [i]), 
                1 + timeout * HZ / 1000);
            sti();
            if (timeLeft == 0) return -EAGAIN;
            if (signal_pending(current)) return -ERESTARTSYS;
        }      
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_ADD_FILTER:
        copy_from_user_ret(&tmpFilterData, (FilterData*) arg, sizeof(FilterData), -EFAULT);
        return addFilter(fileNodePtr, &tmpFilterData);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_CLEAR_FILTERS:
        clearFilters(fileNodePtr);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_FLUSH_RCVBUFFER:
        fileNodePtr -> rcvBufHead = 0;
        fileNodePtr -> rcvBufTail = 0;
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_FLUSH_SENDBUFFER:
        spin_lock(&ctxPtr->sendQLock[fileNodePtr -> chanNr]);
          ctxPtr -> txChanBufHead [fileNodePtr -> chanNr]   = 0;
          ctxPtr -> txChanBufTail [fileNodePtr -> chanNr]   = 0;
          ctxPtr -> sentTXflagCount [fileNodePtr -> chanNr] = 0;
          ctxPtr -> recvTXflagCount [fileNodePtr -> chanNr] = 0;
          ctxPtr -> txQChipFull [fileNodePtr -> chanNr]     = 0;
        spin_unlock(&ctxPtr->sendQLock[fileNodePtr -> chanNr]);

        // empty queue on card
        DEBUG(3 ,"  CMD_FLUSH_QUEUE\n");
        tmpCmd.flushQueue.cmdLen  = (sizeof(cmdFlushQueue)-1) | 0x80;
        tmpCmd.flushQueue.cmdNo   = CMD_FLUSH_QUEUE;
        tmpCmd.flushQueue.channel = fileNodePtr -> chanNr;
        tmpCmd.flushQueue.flags   = FLUSH_TX_QUEUE;
        // should we wait here?
        if (waitTransmit(ctxPtr)) return -1;
        lapcanTransmit(ctxPtr, &tmpCmd);

        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_GET_STAT:
        return lapcanGetStat((CanIfStat*) arg, filp);
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SET_WRITE_BLOCK:
        fileNodePtr -> writeIsBlock = arg;
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SET_READ_BLOCK:
        fileNodePtr -> readIsBlock = arg;
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SET_WRITE_TIMEOUT:
        fileNodePtr -> writeTimeout = arg;
        break;
        /*------------------------------------------------------------------*/
    case LAPCAN_IOC_SET_READ_TIMEOUT:
        fileNodePtr -> readTimeout = arg;
        break;
        /*------------------------------------------------------------------*/
    default:
        return -EINVAL;
    }
    return 0;
}

/*======================================================================*/
/* Process send Q - This function is called from the immediate queue    */
/*======================================================================*/

static void lapcanSend (void *voidLocalPtr)
{
    lpcCmd *p;
    LocalInfo *localPtr;
    LapcanContext *ctxPtr;
    static int chanNr;
    int i;
    localPtr = (LocalInfo*) voidLocalPtr;
    ctxPtr = &(localPtr -> ctx);
    /* Do we have anything to send ? */;
    if (getQLen(ctxPtr -> txCmdBufHead, ctxPtr -> txCmdBufTail, TX_CMD_BUF_SIZE) != 0) {
        DEBUG(8, "Send cmd nr %d from %d\n", 
            getCmdNr(&ctxPtr -> txCmdBuffer [ctxPtr -> txCmdBufTail]),
            ctxPtr -> txCmdBufTail);
        lapcanTransmit(ctxPtr, &ctxPtr -> txCmdBuffer [ctxPtr -> txCmdBufTail]);
        if (++(ctxPtr -> txCmdBufTail) >= TX_CMD_BUF_SIZE)
            ctxPtr -> txCmdBufTail = 0;
        wake_up_interruptible(&ctxPtr -> txCmdWaitQ);
        /* Nothing to do except wait for the card to respond */	  
        return;
    }
    /* Process the channel queues */
    for (i = 0; i < NR_CHANNELS; i++){
        /* Alternate between channels */
        chanNr++;
        if (chanNr >= NR_CHANNELS) chanNr = 0; 
        /* Test if queue is empty or Lapcan has sent "queue high" */
        if (getQLen(ctxPtr -> txChanBufHead [chanNr], ctxPtr -> txChanBufTail [chanNr], 
            TX_CHAN_BUF_SIZE) == 0 || ctxPtr -> txQChipFull [chanNr] ) 
            continue; 
        p = (lpcCmd*) &(ctxPtr -> txChanBuffer [chanNr] [ctxPtr -> txChanBufTail [chanNr]]);
        
        lapcanTransmit(ctxPtr, 
            (lpcCmd*) &(ctxPtr -> txChanBuffer [chanNr] [ctxPtr -> txChanBufTail [chanNr]]));
        DEBUG(8, "Send cmd nr %d from %d:%d\n", chanNr, 
            getCmdNr((lpcCmd*) &ctxPtr -> txChanBuffer [chanNr] 
            [ctxPtr -> txChanBufTail [chanNr]]), ctxPtr -> txChanBufTail [chanNr]);    
        
        /* ??? Multiple CPU race when flushing txQ */
        if (++(ctxPtr -> txChanBufTail [chanNr]) >= TX_CHAN_BUF_SIZE)
            ctxPtr -> txChanBufTail [chanNr] = 0;
        wake_up_interruptible(&ctxPtr -> txChanWaitQ [chanNr]);
        return;
    }  
    /* Oh, this sucks, I can't send anything right now! Oh well, I'll come back later.. */ 
    return;
}

/*======================================================================*/
/* Lapcan interrupt handler                                             */
/*======================================================================*/

static void lapcanInterrupt(int irq, void *dev_id, struct pt_regs *regs)
{
    unsigned short lsrAddr;
    unsigned short fifoAddr;
    unsigned char iir, msr, mcr;
    unsigned char didService;
    int cmdLen, i;
    LocalInfo *local;
    LapcanContext *ctxPtr;
    int maxLoop = 5000;
    int oneFlag = 0;
    
    
    didService = 0;
    local = dev_id;
    ctxPtr = &local -> ctx;
    lsrAddr = ctxPtr -> portAddress+TLH_LSR;
    fifoAddr = ctxPtr -> portAddress+TLH_FIFO;
    
    /* Read interrupt id. register until cleared*/
    while (!((iir = inb(ctxPtr -> portAddress+TLH_IIR))  & IIR_NOINTPEND)) {
        didService = 1;
        //DEBUG(1, "lapcanInterrupt - iir=%x\n",iir);
        maxLoop--;
        if (maxLoop <= 0) {
            DEBUG(1," max_loop\n");
            break;
        }
        
        switch (iir & IIR_INTID_MASK) {
        case INT_RCV_DATA: // Receive standard mode
            //DEBUG(1,"INT_RCV_DATA\n");
            
            if (!(inb(lsrAddr) & LSR_DATA_READY)) {
                oneFlag = 1; // There is one char to read
            }
            
            while (oneFlag || (inb(lsrAddr) & LSR_DATA_READY)) {
                maxLoop--;
                if (maxLoop <= 0) {
                    DEBUG(1," max_loop in standard receive\n");
                    ctxPtr -> curCmdLen = 0;
                    ctxPtr -> curCmdPos = 0;
                    break;
                }
                if (ctxPtr -> curCmdLen == 0) {
                /* We have a new packet incoming...read first byte to be able to determine
                * packet length. The packet is read into a circular buffer(mainRcvBuffer)                                                           
                    */
                    ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead].b [0] = inb(fifoAddr);
                    ctxPtr -> curCmdLen = 
                        getCmdLen(&ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead]);
                    if (ctxPtr -> curCmdLen >= sizeof(lpcCmd)) { // out of sync ?
                        DEBUG(1,"FATAL: Out of sync, data lost\n");
                        ctxPtr -> curCmdLen = 0;
                        ctxPtr -> curCmdPos = 0;
                        break;	   
                    }
                    /* Store in receive buffer at position ctxPtr -> mainRcvBufHead */
                    ctxPtr -> curCmdPos = 1;
                    oneFlag = 0;
                }
                
                // read bytes from fifo
                while (oneFlag || (inb(lsrAddr) & LSR_DATA_READY)) {
                    maxLoop--;
                    if (maxLoop <= 0) {
                        DEBUG(1," max_loop == 0 in lapcanInterrupt (while222)\n");
                        ctxPtr -> curCmdLen = 0;
                        ctxPtr -> curCmdPos = 0;
                        break;
                    }
                    
                    oneFlag = 0;
                    // get data to receive buffer
                    ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead].b[ctxPtr -> curCmdPos] = 
                        inb(fifoAddr);
                    
                    if (++(ctxPtr -> curCmdPos) == ctxPtr -> curCmdLen) {
                        /* We're ready */
                        (ctxPtr -> mainRcvBufHead)++;
                        if (ctxPtr -> mainRcvBufHead >= MAIN_RCV_BUF_SIZE) { 
                            ctxPtr -> mainRcvBufHead = 0;
                        }
                        // Handle the message
                        lapcanReceive(local);
                        ctxPtr -> curCmdLen = 0;
                        break;
                    } // event complete
                    
                    if (ctxPtr -> curCmdPos >= sizeof(lpcCmd)) { // out of synch ?
                        DEBUG(1,"FATAL: Out of sync, data lost\n");
                        ctxPtr -> curCmdLen = 0;
                        ctxPtr -> curCmdPos = 0;
                        break; // restart
                    }
                } // while LSR_DATA_READY
            } // while LSR_DATA_READY
            
            break;
        case INT_TX_HOLD_REG_EMPTY:
            // @@@@ TODO not needed but perhaps conservative?
            //      ctxPtr -> fifoCount = 0;
            //printk("<1>TX_HOLD_REG_EMPTY\n");
            break;
        case INT_MODEM_STATUS:
            /* Read modem status register / clear interrupt */ 
            msr = inb(ctxPtr -> portAddress+TLH_MSR);
            if (msr & MSR_DELTA_DSR) {
                // The subsystem has toggled DSR. i.e. it has read the packet
                // we sent last time. Enable sending a new one.
                ctxPtr -> fifoCount = 0;
                queue_task(&ctxPtr -> txTaskQ, &tq_immediate);      
                mark_bh(IMMEDIATE_BH);
                //lapcanSend(local);
            }
            if ((ctxPtr -> useDCD) && (msr & MSR_DELTA_DCD)) {
                /* New handshake */
                /* Loop. Check if first byte is 0 => packets to read */
                while (!inb(fifoAddr)){
                    ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead].b [0] = inb(fifoAddr);
                    cmdLen = getCmdLen(&ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead]);
                    /* Loop. Read data into buffer */
                    for(i = 1; i < cmdLen; i++){
                        ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufHead].b [i] = inb(fifoAddr);
                    }
                    /* We have read one packet, increase bufferhead */
                    (ctxPtr -> mainRcvBufHead)++;
                    if (ctxPtr -> mainRcvBufHead >= MAIN_RCV_BUF_SIZE) { 
                        ctxPtr -> mainRcvBufHead = 0;
                    }
                    /* Handle message */
                    lapcanReceive(local);
                }
                /* Toggle OUT2 so the card can send more data */
                outb(inb(ctxPtr -> portAddress+TLH_MCR) ^ MCR_OUT2, 
                    ctxPtr -> portAddress+TLH_MCR);
            } 
            break;
    }
    }  
    
    if (didService && !ctxPtr -> useDCD) {
        // reset RTS for auto-RTS feature
        mcr = inb(ctxPtr -> portAddress+TLH_MCR);
        if (!(mcr & MCR_RTS)) {
            outb(mcr |MCR_RTS, ctxPtr -> portAddress+TLH_MCR);
        }
    }
}

/*======================================================================*/
/* Handle incoming lapcan commands                                      */
/*======================================================================*/

static void lapcanReceive (LocalInfo *local)
{
    lpcCmd *cmdPtr;
    LapcanOpenFileNode *fileNodePtr;
    unsigned char *b;
    LapcanContext *ctxPtr;
    unsigned char hasOverrun;
    unsigned char cmdNo = 0;
    int channelNr = 0;
    int rcvQLen;
    struct file *sendId = NULL;
    
    /* Concurrent test */
    if (test_and_set_bit(0, &local -> ctx.recIsBusy))
    {
        DEBUG(1,"REC IS BUSY\n");
        return;
    }
    ctxPtr = &local -> ctx;
    while (ctxPtr -> mainRcvBufHead != ctxPtr -> mainRcvBufTail){
        hasOverrun = 0;
        cmdPtr = &(ctxPtr -> mainRcvBuffer [ctxPtr -> mainRcvBufTail]);
        b = (unsigned char*) cmdPtr;
        /* Check highest bit */
        if (b [0] & 0x80) { /*Infrequent command: */
            cmdNo = getCmdNr(cmdPtr);
            switch(cmdNo) {
            case CMD_CHIP_STATE_EVENT:
                if (cmdPtr -> chipStateEvent.busStatus & BUSSTAT_BUSOFF){
                    ctxPtr -> txQChipFull [cmdPtr -> chipStateEvent.channel] = 0;
                    if (ctxPtr -> fifoCount == 0) {
                        queue_task(&ctxPtr -> txTaskQ, &tq_immediate);
                        mark_bh(IMMEDIATE_BH);
                    }
                    ctxPtr -> sentTXflagCount [cmdPtr -> chipStateEvent.channel] = 0;
                    ctxPtr -> recvTXflagCount [cmdPtr -> chipStateEvent.channel] = 0;
                    
                }
                DEBUG (4,"CMD_CHIP_STATE_EVENT channel=%d, busStatus=%d,\
                    txerr=%d, rxerr=%d , time=%d\n", cmdPtr -> chipStateEvent.channel,
                    cmdPtr -> chipStateEvent.busStatus,  cmdPtr -> chipStateEvent.txErrorCounter,
                    cmdPtr -> chipStateEvent.rxErrorCounter,  cmdPtr -> chipStateEvent.time);
                break;
            case CMD_GET_CHIP_STATE_RESP:
            DEBUG (5,"CMD_CHIP_STATE_RESP channel=%d, busStatus=%d,\
                txerr=%d, rxerr=%d\n", cmdPtr -> getChipStateResp.channel,
                cmdPtr -> getChipStateResp.busStatus,  cmdPtr -> getChipStateResp.txErrorCounter,
                cmdPtr -> getChipStateResp.rxErrorCounter);
            
            case CMD_GET_BUSPARAMS_RESP:
            DEBUG (7,"BUSPARAMS RESP chan=%u, bitRate=%lu, tseg1=%u, tseg2=%u, sjw=%u,\
                nosamp=%d\n", 
                cmdPtr -> getBusparamsResp.channel, cmdPtr -> getBusparamsResp.bitRate, 
                cmdPtr -> getBusparamsResp.tseg1, cmdPtr -> getBusparamsResp.tseg2, 
                cmdPtr -> getBusparamsResp.sjw, cmdPtr -> getBusparamsResp.noSamp);
            break;
            case CMD_GET_CARD_INFO_RESP:
                ctxPtr -> hwRevision = cmdPtr -> getCardInfoResp.hwInfo & 0x0f; // Revision;
                ctxPtr -> clockResolution = cmdPtr -> getCardInfoResp.clockResolution;
                ctxPtr -> hwInfoIsValid = 1;
                break;
            case CMD_GET_SOFTWARE_INFO_RESP:
                ctxPtr -> swRevision =  (cmdPtr -> getSoftwareInfoResp.applicationVersion [0] - '0') << 20;
                ctxPtr -> swRevision += (cmdPtr -> getSoftwareInfoResp.applicationVersion [1] - '0') << 16;
                ctxPtr -> swRevision += (cmdPtr -> getSoftwareInfoResp.applicationVersion [2] - '0') << 12;
                ctxPtr -> swRevision += (cmdPtr -> getSoftwareInfoResp.applicationVersion [3] - '0') <<  8;
                ctxPtr -> swRevision += (cmdPtr -> getSoftwareInfoResp.applicationVersion [4] - '0') <<  4;
                ctxPtr -> swRevision += (cmdPtr -> getSoftwareInfoResp.applicationVersion [5] - '0');
                if ((cmdPtr -> b [0] & 0x7F) > offsetof(cmdGetSoftwareInfoResp, swOptions)-1) {
                    ctxPtr -> swOptions = cmdPtr -> getSoftwareInfoResp.swOptions;
                    DEBUG (5, "swOptions=%u\n", cmdPtr -> getSoftwareInfoResp.swOptions);
                } else {
                    // There's no swOptions field! (The firmware is too old.)
                    ctxPtr -> swOptions = 0;
                }
                ctxPtr -> swInfoIsValid = 1;
                break;
            case CMD_ERROR_EVENT:
                if (cmdPtr -> errorEvent.errorCode == LAPERR_QUEUE_LEVEL){
                    channelNr = cmdPtr -> errorEvent.addInfo1;
                    if (cmdPtr -> errorEvent.addInfo2 == TRANSMIT_QUEUE_HIGH){	    
                        ctxPtr -> txQChipFull [channelNr] = 1;	    
                        //	    DEBUG(1,"TRANSMIT QUEUE IS HIGH!\n");
                    } 
                    else if (cmdPtr -> errorEvent.addInfo2 == TRANSMIT_QUEUE_LOW){
                        if (ctxPtr -> txQChipFull [channelNr]){
                            /* We previously had a full queue, try to send again */ 
                            // DEBUG(1,"TRANSMIT QUEUE IS LOW!\n");
                            if (ctxPtr -> fifoCount == 0) {
                                queue_task(&ctxPtr -> txTaskQ, &tq_immediate);
                                mark_bh(IMMEDIATE_BH);
                            }
                        }
                        ctxPtr -> txQChipFull [channelNr] = 0;	    
                    }
                }
                break;
            default:
                DEBUG(8,"InFreq Msg cmd=%d, ptr=%d, chan =%d\n", cmdNo, ctxPtr -> mainRcvBufTail
                    , getCmdChan(cmdPtr));
                break;
            }
        } else {
            /* Extract information */
            channelNr = getCmdChan(cmdPtr);
            //      DEBUG(8,"Freq Msg cmd=%d, len=%d, chan=%d, dlc=%d, flags=%d\n", 
            //	    getCmdNr(cmdPtr), getCmdLen(cmdPtr), getCmdChan(cmdPtr), b [1] & 0x0f, b[2]);
            hasOverrun = getFlags(cmdPtr) & MSGFLAG_OVERRUN;
            /* Is this an echoed message ? */
            if (getFlags(cmdPtr) & MSGFLAG_TX) {
                ctxPtr -> recvTXflagCount [channelNr]++;
                if (ctxPtr -> sentTXflagCount [channelNr] == ctxPtr -> recvTXflagCount [channelNr])
                    /* Wake up those who are waiting for all sending to finish */
                    wake_up_interruptible (&ctxPtr -> flushQ [channelNr]);
            } else {
                sendId = NULL;
            }
        }
        if (++ctxPtr -> mainRcvBufTail >= MAIN_RCV_BUF_SIZE) { 
            ctxPtr -> mainRcvBufTail = 0;
        }
        /* Update and notify readers */
        fileNodePtr = ctxPtr -> openFileList;
        for (; fileNodePtr != NULL; fileNodePtr = fileNodePtr -> next){
            /* ??? FILE_RCV_BUF_SIZE - 1 */
            if (checkFilters(fileNodePtr, cmdPtr)) {
                rcvQLen = getQLen(fileNodePtr -> rcvBufHead, 
                    fileNodePtr -> rcvBufTail, FILE_RCV_BUF_SIZE);
                if (rcvQLen >= FILE_RCV_BUF_SIZE - 1){
                    /* The buffer is full, ignore new message */
                    fileNodePtr -> overruns [channelNr]++;
                    /* Mark message that arrived before this one */
                    if (fileNodePtr -> rcvBufHead == 0){
                        setFlags(&(fileNodePtr -> fileRcvBuffer [FILE_RCV_BUF_SIZE - 1]), 
                            MSGFLAG_OVERRUN | 
                            getFlags(&(fileNodePtr -> fileRcvBuffer [FILE_RCV_BUF_SIZE - 1])));
                    } else {
                        setFlags(&(fileNodePtr -> fileRcvBuffer [fileNodePtr -> rcvBufHead - 1]), 
                            MSGFLAG_OVERRUN |
                            getFlags(&(fileNodePtr -> fileRcvBuffer [fileNodePtr -> rcvBufHead - 1]))); 
                    }
                } else {
                    //DEBUG(1, "copy msg with id = %d, to %d, filp = %p\n",getCmdNr(cmdPtr),
                    //	fileNodePtr -> rcvBufHead, fileNodePtr -> filp);
                    copyLapcanCmd(&(fileNodePtr -> fileRcvBuffer [fileNodePtr -> rcvBufHead]), cmdPtr);
                    if (hasOverrun) (fileNodePtr -> overruns [channelNr])++;
                    /* Increase and wakeup reader */
                    (fileNodePtr -> rcvBufHead)++;
                    if (fileNodePtr -> rcvBufHead >= FILE_RCV_BUF_SIZE) 
                        fileNodePtr -> rcvBufHead = 0;
                    if (rcvQLen == 0) wake_up_interruptible(&fileNodePtr -> rxWaitQ);
                } 
            } else 
                ;
            //DEBUG(1, "BLOCKED  msg with id = %d, to %d filp = %p \n",getCmdNr(cmdPtr),
            //	fileNodePtr -> rcvBufHead, fileNodePtr -> filp);
        }
    }
    clear_bit(0, &local -> ctx.recIsBusy);
    return;
}

/*======================================================================*/
/*    Check for LAPcan presence                                         */
/*======================================================================*/

static int lapcanVerify(LocalInfo *localPtr)
{
    LapcanContext *ctxPtr;
    ctxPtr = &localPtr -> ctx; 
    if ((inb(ctxPtr -> portAddress+2) & 0x30) != 0 || // Bit 2.4,2.5 should be zero
        (inb(ctxPtr -> portAddress+4) & 0xc0) != 0)
    {
        ctxPtr -> isPresent = 0;
        return -1;
    } else {
        ctxPtr -> isPresent = 1;
        return 0;
    }
    
}

/*======================================================================*/
/*    Init datastructures                                               */
/*======================================================================*/

static int lapcanInitContext(LocalInfo *localPtr)
{
    int i;
    LapcanContext *ctxPtr;
    ctxPtr = &localPtr -> ctx;
    ctxPtr -> portAddress = (localPtr->link).io.BasePort1;
    ctxPtr -> txTaskQ.routine = &lapcanSend;
    ctxPtr -> txTaskQ.data = localPtr;
    /* Init waitqueues */
    init_waitqueue_head (&(ctxPtr -> txCmdWaitQ));
    for(i = 0; i < NR_CHANNELS; i++) {
        init_waitqueue_head (&(ctxPtr -> flushQ[i]));
        init_waitqueue_head (&(ctxPtr -> txChanWaitQ[i]));
        spin_lock_init(&(ctxPtr -> exclAndOpenLock[i]));
        spin_lock_init(&(ctxPtr -> sendQLock[i]));
        
    }
    
    return 0;
}
/*======================================================================*/
/*  Initialize the card                                                 */
/*======================================================================*/

static int switchHandshake (LapcanContext *ctxPtr) 
{
    lpcCmd cmd;
    cmd.setGlobalOptionsReq.cmdLen = (sizeof(cmdSetGlobalOptionsReq)-1) | 0x80;
    cmd.setGlobalOptionsReq.cmdNo  = CMD_SET_GLOBAL_OPTIONS_REQ;
    cmd.setGlobalOptionsReq.optionsMask = GLOPT_USE_DCD_PROTOCOL;
    cmd.setGlobalOptionsReq.optionsCode = GLOPT_USE_DCD_PROTOCOL;
    lapcanTransmit(ctxPtr, &cmd);
    if (waitTransmit(ctxPtr) == 0) {
        ctxPtr -> useDCD = 1;
        ctxPtr -> useOUT1Toggle = 1;
        /* Enable only modem status interrupts */
        outb(IER_EDSSI, ctxPtr -> portAddress+TLH_IER);
        return 0;
    } else {
        return -1;
    }
}

/*======================================================================*/
/*  Wait until fifo is avalable                                         */
/*======================================================================*/

static int waitTransmit (LapcanContext *ctxPtr) 
{
    /* jiffies equals nr of timer interrupts since boot */
    long int timeOut = jiffies + 1 * HZ;
    for (;;) {
        if (ctxPtr -> fifoCount == 0) return 0;
        if (jiffies >= timeOut) break;
    }
    return -1;
}

/*======================================================================*/
/*  Init card & start firmware application                              */
/*======================================================================*/

static int lapcanInit (LapcanContext *ctxPtr)
{
    lpcCmd cmd;
    long int timeOut;
    int error;
    int i;
    unsigned char transId = 0; 
    
    if (ctxPtr -> isPresent) {
        
        // There is no way to reset the card, so we start directly by
        // setting up the FIFO.
        
        outb(0x00,ctxPtr -> portAddress+TLH_FCR);              // FIFO disable
        outb(FCR_FIFO_ENABLE,ctxPtr -> portAddress+TLH_FCR);   // FIFO enable
        
        // Assign the Divisor Latch
        outb(LCR_DLAB,ctxPtr -> portAddress+TLH_LCR);
        outb(64,ctxPtr -> portAddress+TLH_DLL);
        outb(0,ctxPtr -> portAddress+TLH_DLM);
        
        // Clear the DLAB-bit to get access to the FIFO
        outb(0,ctxPtr -> portAddress+TLH_LCR); // Clear bit LCR_DLAB
        // Enable and reset the FIFOs (for 64 byte operation), trigger level 1 byte
        outb(
            FCR_FIFO_ENABLE|    // FIFO Enable
            FCR_RXFIFO_RESET|   // FIFO Reset
            FCR_TXFIFO_RESET|
            FCR_AUTO_RTS|       // auto-RTS for '564
            FCR_FDEPTH64,ctxPtr -> portAddress+TLH_FCR);      // 64 bytes FIFO, triggerlevel 1
        
        // Assert RTS# and OUT1#.
        outb(MCR_RTS,ctxPtr -> portAddress+TLH_MCR);
        
        // Enable all Interrupts
        outb(IER_EDSSI | IER_ERBI | IER_ETBEI | IER_ELSI,ctxPtr -> portAddress+TLH_IER);
        
        // for security empty the FIFO, to get an interrupt
        DEBUG(7," empty fifo\n");
        while ((inb(ctxPtr -> portAddress+TLH_LSR) & 1)) {
            inb(ctxPtr -> portAddress+TLH_FIFO);
        }
        DEBUG(7," ok\n");
        
        /* Get some information */
        
        if (ctxPtr -> isInit) return 0;
        
        cmd.startApplicationReq.cmdLen = (sizeof(cmdStartApplicationReq)-1) | 0x80;
        cmd.startApplicationReq.cmdNo  = CMD_START_APPLICATION_REQ;
        cmd.startApplicationReq.application = APP_SHUFFLE;
        cmd.startApplicationReq.startAddress = 0L;
        cmd.startApplicationReq.p1 = 0;
        cmd.startApplicationReq.p2 = 0;
        lapcanTransmit(ctxPtr, &cmd);
        
        DEBUG(4," transmit startapp\n");
        
        /* Wait and let other processes run for 1/2 second*/
        current -> state = TASK_INTERRUPTIBLE;
        schedule_timeout(0.5 * HZ); 
        
        cmd.getCardInfoReq.cmdLen    = (sizeof(cmdGetCardInfoReq)-1) | 0x80;
        cmd.getCardInfoReq.cmdNo     = CMD_GET_CARD_INFO_REQ;
        cmd.getCardInfoReq.transId   = transId++;
        cmd.getCardInfoReq.dataLevel = 0; // qqq
        lapcanTransmit(ctxPtr, &cmd);
        if (waitTransmit(ctxPtr)) return -1;  
        lapcanTransmit(ctxPtr, &cmd);
        
        cmd.getSoftwareInfoReq.cmdLen  = (sizeof(cmdGetSoftwareInfoReq)-1) | 0x80;
        cmd.getSoftwareInfoReq.cmdNo   = CMD_GET_SOFTWARE_INFO_REQ;
        cmd.getSoftwareInfoReq.transId = transId++;
        if (waitTransmit(ctxPtr)) return -1;  
        lapcanTransmit(ctxPtr, &cmd);
        
        error = -1;
        timeOut = jiffies + 1 * HZ; // 1 sec ; the debug f/w needs some time to answer.
        for (;;) {
            if ((volatile int)(ctxPtr -> hwInfoIsValid) &&
                (volatile int)(ctxPtr -> swInfoIsValid)) {
                error = 0;
                break;
            }
            schedule();
            if (jiffies >= timeOut) break;
        }
        if (error) {
            DEBUG(1," Warning - Not initialized !!!!!!!!\n");
            return error;
        }
        
        DEBUG(1,"  hwInfoIsValid = %u\n",ctxPtr -> hwInfoIsValid);
        DEBUG(1,"  swInfoIsValid = %u\n",ctxPtr -> swInfoIsValid);
        DEBUG(1,"  hwRevision = %u\n",ctxPtr -> hwRevision);
        DEBUG(1,"  swRevision = %u\n",ctxPtr -> swRevision);
        DEBUG(1,"  swOptions = %u\n",ctxPtr -> swOptions);
        DEBUG(1,"  clockResolution = %u\n",ctxPtr -> clockResolution);
        
        // Use the DCD handshaking protocol
        if (ctxPtr -> swOptions & SWOPT_CAN_TOGGLE_DCD) 
        {
            if (switchHandshake(ctxPtr) != 0) return -1; /* OK? */
        }
        
        DEBUG(1,"useDCD = %u\n",ctxPtr -> useDCD);
        DEBUG(1,"useOUT1Toggle = %u\n",ctxPtr -> useOUT1Toggle);
        
        if (!(ctxPtr -> swOptions & SWOPT_CAN_TOGGLE_DCD)) {
            printk("LAPCAN WARNING: Old firmware, please upgrade\n");
        }
        
        for (i=0;i<=1;i++) {
            DEBUG(3,"  CMD_SET_DRIVERMODE channel=%d\n",i);
            cmd.setDrivermodeReq.cmdLen = (sizeof(cmdSetDrivermodeReq)-1) | 0x80;
            cmd.setDrivermodeReq.cmdNo  = CMD_SET_DRIVERMODE_REQ;
            cmd.setDrivermodeReq.transId = transId++;
            cmd.setDrivermodeReq.channel = (unsigned char) i;
            cmd.setDrivermodeReq.driverMode = DRIVERMODE_NORMAL;
            if (waitTransmit(ctxPtr)) return -1;
            lapcanTransmit(ctxPtr, &cmd);
        }
        
        
        for (i=0;i<=1;i++) {
            DEBUG(3,"  CMD_RESET_CHIP channel=%d\n",i);
            cmd.resetChipReq.cmdLen = (sizeof(cmdResetChipReq)-1) | 0x80;
            cmd.resetChipReq.cmdNo  = CMD_RESET_CHIP_REQ;
            cmd.resetChipReq.transId = transId++;
            cmd.resetChipReq.channel = (unsigned char) i;
            if (waitTransmit(ctxPtr)) return -1;
            lapcanTransmit(ctxPtr, &cmd);
        }
        ctxPtr -> isInit = 1;  
        
        return 0;
    } else {
        return -1; /* No card found */
    }
}

/*======================================================================*/
/* Send a lapcan command to the card                                    */
/*======================================================================*/

static void lapcanTransmit (LapcanContext *ctxPtr, lpcCmd *msg)
{
    unsigned int len;
    unsigned int i;
    unsigned char *p = msg -> b;
    unsigned char mcr;
    unsigned short addr = ctxPtr -> portAddress+TLH_FIFO;
    
    if (ctxPtr -> fifoCount != 0) return; /* We shouldn't be here! */
    ctxPtr -> fifoCount = 1;
    len = getCmdLen(msg);
    
    for (i = 0; i < len; i++) {
        outb(p [i],addr);
    }
    
    // Tell the card there's something to read.
    cli(); /* We dont want any interrupts while toggling */
    mcr = inb(ctxPtr -> portAddress+TLH_MCR);
    if (ctxPtr -> useOUT1Toggle) {
        // Toggle OUT1#.
        outb((unsigned char)(mcr ^ MCR_OUT1),ctxPtr -> portAddress+TLH_MCR);
    } 
    else {
        // Pulse OUT1#.
        mcr &= ~MCR_OUT1;
        // Write MCR twice to cope with platforms where the I/O cycles
        // are relatively fast. The CANcardX needs at least 500 ns
        // to safely detect a level change.
        outb((unsigned char )(mcr | MCR_OUT1),ctxPtr -> portAddress+TLH_MCR);
        outb((unsigned char )(mcr | MCR_OUT1),ctxPtr -> portAddress+TLH_MCR);
        outb(mcr,ctxPtr -> portAddress+TLH_MCR);
    }
    sti();  
    return;
} 

/*======================================================================*/
/* Report card services errors                                          */
/*======================================================================*/

static void cs_error(client_handle_t handle, int func, int ret)
{
    error_info_t err = { func, ret };
    CardServices(ReportError, handle, &err);
}

/*======================================================================*/
/*lapcanAttach() creates an "instance" of the driver, allocating        */
/*local data structures for one device.  The device is registered       */
/*with Card Services.                                                   */
/*                                                                      */
/*The dev_link structure is initialized, but we don't actually          */
/*configure the card at this point -- we wait until we receive a        */
/*  card insertion event.                                               */
/*======================================================================*/

static dev_link_t *lapcanAttach(void)
{
    LocalInfo *local;
    dev_link_t *link;
    client_reg_t client_reg;
    int ret, i;
    
    DEBUG(6, "lapcan_attach!()\n");
    
    /* Allocate space for private device-specific data */
    local = kmalloc(sizeof(LocalInfo), GFP_KERNEL);
    if (!local) return NULL;
    memset(local, 0, sizeof(LocalInfo));
    
    link = &local -> link; link -> priv = local; 
    /* We want to irq.Instance in the ISR */
    link -> irq.Instance = local;
    
    /* Initialize the dev_link_t structure */
    link -> release.function = &lapcanRelease;
    link -> release.data = (u_long)link;
    
    /* Interrupt setup */
    link -> irq.Attributes = IRQ_TYPE_EXCLUSIVE | IRQ_HANDLE_PRESENT;
    link -> irq.IRQInfo1 = IRQ_INFO2_VALID|IRQ_LEVEL_ID;
    if (irq_list [0] == -1)
        link -> irq.IRQInfo2 = irq_mask;
    else
        for (i = 0; i < 4; i++)
            link -> irq.IRQInfo2 |= 1 << irq_list [i];
        link -> irq.Handler = &lapcanInterrupt;
        
        /*
        General socket configuration defaults can go here.  In this
        client, we assume very little, and rely on the CIS for almost
        everything.  In most clients, many details (i.e., number, sizes,
        and attributes of IO windows) are fixed by the nature of the
        device, and can be hard-wired here.
        */
        link -> conf.Attributes = 0;
        link -> conf.Vcc = 50;
        link -> conf.IntType = INT_MEMORY_AND_IO;
        
        /* Register with Card Services */
        link -> next = devList;
        devList = link;
        client_reg.dev_info = &devInfo;
        client_reg.Attributes = INFO_IO_CLIENT | INFO_CARD_SHARE;
        client_reg.EventMask =
            CS_EVENT_CARD_INSERTION | CS_EVENT_CARD_REMOVAL |
            CS_EVENT_RESET_PHYSICAL | CS_EVENT_CARD_RESET |
            CS_EVENT_PM_SUSPEND | CS_EVENT_PM_RESUME;
        client_reg.event_handler = &lapcanEvent;
        client_reg.Version = 0x0210;
        client_reg.event_callback_args.client_data = link;
        ret = CardServices(RegisterClient, &link -> handle, &client_reg);
        if (ret != CS_SUCCESS) {
            cs_error(link -> handle, RegisterClient, ret);
            lapcanDetach(link);
            return NULL;
        }
        
        return link;
} /* lapcanAttach */

/*======================================================================*/
/*    This deletes a driver "instance".  The device is de-registered    */
/*    with Card Services.  If it has been released, all local data      */
/*    structures are freed.  Otherwise, the structures will be freed    */
/*    when the device is released.                                      */
/*======================================================================*/

static void lapcanDetach(dev_link_t *link)
{
    dev_link_t **linkp;
    
    DEBUG(6, "lapcanDetach(0x%p)\n", link);
    
    /* Locate device structure */
    for (linkp = &devList; *linkp; linkp = &(*linkp) -> next)
        if (*linkp == link) break;
        if (*linkp == NULL)
            return;
        
            /*
            If the device is currently configured and active, we won't
            actually delete it yet.  Instead, it is marked so that when
            the release() function is called, that will trigger a proper
            detach().
        */
        if (link -> state & DEV_CONFIG) {
#ifdef PCMCIA_DEBUG
            printk(KERN_DEBUG "lapcan_cs: detach postponed, '%s' "
                "still locked\n", link -> dev -> dev_name);
#endif
            link -> state |= DEV_STALE_LINK;
            return;
        }
        
        /* Unregister device */
        
        
        /* Break the link with Card Services */
        if (link -> handle)
            CardServices(DeregisterClient, link -> handle);
        
        /* Unlink device structure, and free it */
        *linkp = link -> next;
        /* This points to the parent LocalInfo struct */
        kfree(link -> priv);
        
} /* lapcanDetach */

/*======================================================================*/
/* lapcanConfig() is scheduled to run after a CARD_INSERTION event      */
/* is received, to configure the PCMCIA socket, and to make the         */
/* device available to the system.                                      */
/*======================================================================*/

#define CS_CHECK(fn, args...) \
while ((last_ret=CardServices(last_fn=(fn),args))!=0) goto cs_failed

#define CFG_CHECK(fn, args...) \
if (CardServices(fn, args) != 0) goto next_entry

static void lapcanConfig(dev_link_t *link)
{
    client_handle_t handle = link -> handle;
    LocalInfo *dev = link -> priv;
    LocalInfo *localPtr;
    tuple_t tuple;
    cisparse_t parse;
    int last_fn, last_ret;
    u_char buf[64];
    config_info_t conf;
    win_req_t req;
    memreq_t map;
    unsigned char minor;
    
    DEBUG(6, "lapcanConfig(0x%p)\n", link);
    /*
    This reads the card's CONFIG tuple to find its configuration
    registers.
    */
    tuple.DesiredTuple = CISTPL_CONFIG;
    tuple.Attributes = 0;
    tuple.TupleData = buf;
    tuple.TupleDataMax = sizeof(buf);
    tuple.TupleOffset = 0;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    CS_CHECK(GetTupleData, handle, &tuple);
    CS_CHECK(ParseTuple, handle, &tuple, &parse);
    link -> conf.ConfigBase = parse.config.base;
    link -> conf.Present = parse.config.rmask [0];
    
    /* Configure card */
    link -> state |= DEV_CONFIG;
    
    /* Look up the current Vcc */
    CS_CHECK(GetConfigurationInfo, handle, &conf);
    link -> conf.Vcc = conf.Vcc;
    
    /*
    In this loop, we scan the CIS for configuration table entries,
    each of which describes a valid card configuration, including
    voltage, IO window, memory window, and interrupt settings.
    */
    
    tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
    CS_CHECK(GetFirstTuple, handle, &tuple);
    
    while (1) {
        cistpl_cftable_entry_t dflt = { 0 };
        cistpl_cftable_entry_t *cfg = &(parse.cftable_entry);
        CFG_CHECK(GetTupleData, handle, &tuple);
        CFG_CHECK(ParseTuple, handle, &tuple, &parse);
        
        if (cfg -> flags & CISTPL_CFTABLE_DEFAULT) dflt = *cfg;
        if (cfg -> index == 0) goto next_entry;
        link -> conf.ConfigIndex = cfg -> index;
        
        /* Does this card need audio output? */
        if (cfg -> flags & CISTPL_CFTABLE_AUDIO) {
            link -> conf.Attributes |= CONF_ENABLE_SPKR;
            link -> conf.Status = CCSR_AUDIO_ENA;
        }
        
        /* Use power settings for Vcc and Vpp if present */
        /*  Note that the CIS values need to be rescaled */
        if (cfg -> vcc.present & (1<<CISTPL_POWER_VNOM)) {
            if (conf.Vcc != cfg -> vcc.param [CISTPL_POWER_VNOM] / 10000)
                goto next_entry;
        } else if (dflt.vcc.present & (1<<CISTPL_POWER_VNOM)) {
            if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]/10000)
                goto next_entry;
        }
        
        if (cfg -> vpp1.present & (1<<CISTPL_POWER_VNOM))
            link -> conf.Vpp1 = link -> conf.Vpp2 =
            cfg -> vpp1.param[CISTPL_POWER_VNOM]/10000;
        else if (dflt.vpp1.present & (1<<CISTPL_POWER_VNOM))
            link -> conf.Vpp1 = link -> conf.Vpp2 =
            dflt.vpp1.param[CISTPL_POWER_VNOM]/10000;
        
        /* Do we need to allocate an interrupt? */
        if (cfg -> irq.IRQInfo1 || dflt.irq.IRQInfo1)
            link -> conf.Attributes |= CONF_ENABLE_IRQ;
        
        /* IO window settings */
        link -> io.NumPorts1 = link -> io.NumPorts2 = 0;
        if ((cfg -> io.nwin > 0) || (dflt.io.nwin > 0)) {
            cistpl_io_t *io = (cfg -> io.nwin) ? &cfg -> io : &dflt.io;
            link -> io.Attributes1 = IO_DATA_PATH_WIDTH_AUTO;
            if (!(io -> flags & CISTPL_IO_8BIT))
                link -> io.Attributes1 = IO_DATA_PATH_WIDTH_16;
            if (!(io -> flags & CISTPL_IO_16BIT))
                link -> io.Attributes1 = IO_DATA_PATH_WIDTH_8;
            link -> io.IOAddrLines = io -> flags & CISTPL_IO_LINES_MASK;
            link -> io.BasePort1 = io -> win[0].base;
            link -> io.NumPorts1 = io -> win[0].len;
            if (io -> nwin > 1) {
                link -> io.Attributes2 = link -> io.Attributes1;
                link -> io.BasePort2 = io -> win[1].base;
                link -> io.NumPorts2 = io -> win[1].len;
            }
        }
        
        /* This reserves IO space but doesn't actually enable it */
        CFG_CHECK(RequestIO, link -> handle, &link -> io);
        /*
        Now set up a common memory window, if needed.  There is room
        in the dev_link_t structure for one memory window handle,
        but if the base addresses need to be saved, or if multiple
        windows are needed, the info should go in the private data
        structure for this device.
        
          Note that the memory window base is a physical address, and
          needs to be mapped to virtual space with ioremap() before it
          is used.
        */
        if ((cfg -> mem.nwin > 0) || (dflt.mem.nwin > 0)) {
            cistpl_mem_t *mem =
                (cfg -> mem.nwin) ? &cfg -> mem : &dflt.mem;
            req.Attributes = WIN_DATA_WIDTH_16|WIN_MEMORY_TYPE_CM;
            req.Attributes |= WIN_ENABLE;
            req.Base = mem -> win[0].host_addr;
            req.Size = mem -> win[0].len;
            req.AccessSpeed = 0;
            link -> win = (window_handle_t)link -> handle;
            CFG_CHECK(RequestWindow, &link -> win, &req);
            map.Page = 0; map.CardOffset = mem -> win[0].card_addr;
            CFG_CHECK(MapMemPage, link -> win, &map);
        }
        /* If we got this far, we're cool! */
        break;
        
next_entry:
        CS_CHECK(GetNextTuple, handle, &tuple);
    }
    
    /*
    Allocate an interrupt line.  Note that this does not assign a
    handler to the interrupt, unless the 'Handler' member of the
    irq structure is initialized.
    */
    if (link -> conf.Attributes & CONF_ENABLE_IRQ) 
        CS_CHECK(RequestIRQ, link -> handle, &link -> irq);
    
        /*
        This actually configures the PCMCIA socket -- setting up
        the I/O windows and the interrupt mapping, and putting the
        card and host interface into "Memory and IO" mode.
    */
    CS_CHECK(RequestConfiguration, link -> handle, &link -> conf);
    
    /*
    We can release the IO port allocations here, if some other
    driver for the card is going to loaded, and will expect the
    ports to be available.
    */
    if (free_ports) {
        if (link -> io.BasePort1)
            release_region(link -> io.BasePort1, link -> io.NumPorts1);
        if (link -> io.BasePort2)
            release_region(link -> io.BasePort2, link -> io.NumPorts2);
    }
    
    /*
    At this point, the dev_node_t structure(s) need to be
    initialized and arranged in a linked list at link -> dev.
    */
    
    /* Find first free minor number */
    
    for (minor = 0; minor < 255; minor++) {
        localPtr = devList -> priv;
        while (localPtr != NULL && (localPtr -> node1.minor != minor || localPtr == dev)) {
            if (localPtr -> link.next == NULL) {
                localPtr = NULL;
            } 
            else {
                localPtr = localPtr -> link.next -> priv;
            }
        }
        if (localPtr == NULL) {
            break;
        }	
    }
    
    /* We couldn't find any free minor number!!!!!!!*/
    if (minor == 255) goto cs_failed;
    
    DEBUG(7,"minor=%d\n", minor);
    sprintf(dev -> node1.dev_name, "lapcan%d", minor); /*MBUG*/
    dev -> node1.major = CAN_MAJOR; 
    dev -> node1.minor = minor;
    link -> dev = &dev -> node1;
    dev -> node1.next = NULL;
    
    /* Time to initialize the hardware*/
    
    lapcanInitContext(dev);
    if (lapcanVerify(dev) != 0) goto cs_failed;
    if (lapcanInit(&dev -> ctx) != 0) goto cs_failed;
    
    /* Finally, report what we've done */
    printk(KERN_INFO "%s: index 0x%02x: Vcc %d.%d",
        dev -> node1.dev_name, link -> conf.ConfigIndex,
        link -> conf.Vcc/10, link -> conf.Vcc%10);
    if (link -> conf.Vpp1)
        printk(", Vpp %d.%d", link -> conf.Vpp1/10, link -> conf.Vpp1%10);
    if (link -> conf.Attributes & CONF_ENABLE_IRQ)
        printk(", irq %d", link -> irq.AssignedIRQ);
    if (link -> io.NumPorts1)
        printk(", io 0x%04x-0x%04x", link -> io.BasePort1,
	       link -> io.BasePort1+link -> io.NumPorts1-1);
    if (link -> io.NumPorts2)
        printk(" & 0x%04x-0x%04x", link -> io.BasePort2,
	       link -> io.BasePort2+link -> io.NumPorts2-1);
    if (link -> win)
        printk(", mem 0x%06lx-0x%06lx", req.Base,
	       req.Base+req.Size-1);
    link -> state &= ~DEV_CONFIG_PENDING;
    
    return;
    
cs_failed:
    DEBUG(2,"lapcanConfig failed\n");
    cs_error(link -> handle, last_fn, last_ret);
    lapcanRelease((u_long)link);
    
} /* lapcanConfig */

/*======================================================================*/
/* After a card is removed, lapcanRelease() will unregister the         */
/* device, and release the PCMCIA configuration.  If the device is      */
/* still open, this will be postponed until it is closed.               */
/*======================================================================*/

static void lapcanRelease(u_long arg)
{
    dev_link_t *link = (dev_link_t *)arg;
    
    DEBUG(6, "lapcanRelease(0x%p)\n", link);
    
    /*
    If the device is currently in use, we won't release until it
    is actually closed, because until then, we can't be sure that
    no one will try to access the device or its data structures.
    */
    if (link -> open) {
        DEBUG(6, "lapcan_cs: release postponed, '%s' still open\n",
            link -> dev -> dev_name);
        link -> state |= DEV_STALE_CONFIG;
        return;
    }
    
    /* Unlink the device chain */
    link -> dev = NULL;
    if (link -> win)
        CardServices(ReleaseWindow, link -> win);
    CardServices(ReleaseConfiguration, link -> handle);
    if (link -> io.NumPorts1)
        CardServices(ReleaseIO, link -> handle, &link -> io);
    if (link -> irq.AssignedIRQ)
        CardServices(ReleaseIRQ, link -> handle, &link -> irq);
    link -> state &= ~DEV_CONFIG;
    
    if (link -> state & DEV_STALE_LINK)
        lapcanDetach(link);
    
} /* lapcanRelease */

/*======================================================================*/
/* The card status event handler.  Mostly, this schedules other         */
/* stuff to run after an event is received.                             */
/*                                                                      */
/* When a CARD_REMOVAL event is received, we immediately set a          */
/* private flag to block future accesses to this device.  All the       */
/* functions that actually access the device should check this flag     */
/* to make sure the card is still present.                              */
/*======================================================================*/

static int lapcanEvent(event_t event, int priority,
                       event_callback_args_t *args)
{
    dev_link_t *link = args -> client_data;
    LocalInfo *dev = link -> priv;
    DEBUG(4, "lapcanEvent(0x%06x)\n", event);
    switch (event) {
    case CS_EVENT_CARD_REMOVAL:
        link -> state &= ~DEV_PRESENT;
        if (link -> state & DEV_CONFIG) {
            ((LocalInfo *)link -> priv) -> stop = 1;
            link -> release.expires = jiffies + HZ/20;
            add_timer(&link -> release);
        }
        break;
    case CS_EVENT_CARD_INSERTION:
        link -> state |= DEV_PRESENT | DEV_CONFIG_PENDING;
        dev -> bus = args -> bus;
        lapcanConfig(link);
        break;
    case CS_EVENT_PM_SUSPEND:
        link -> state |= DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_RESET_PHYSICAL:
        /* Mark the device as stopped, to block IO until later */
        dev -> stop = 1;
        if (link -> state & DEV_CONFIG)
            CardServices(ReleaseConfiguration, link -> handle);
        break;
    case CS_EVENT_PM_RESUME:
        link -> state &= ~DEV_SUSPEND;
        /* Fall through... */
    case CS_EVENT_CARD_RESET:
        if (link -> state & DEV_CONFIG)
            CardServices(RequestConfiguration, link -> handle, &link -> conf);
        dev -> stop = 0;
        /* We need to restart the card */
        break;
    }
    return 0;
} /* lapcanEvent */

/*======================================================================*/
/* Run when driver is loaded                                            */
/*======================================================================*/

int init_module(void)
{
    servinfo_t serv;
    int error;
    DEBUG(1,"%s\n", version);
    DEBUG(1, "< "__DATE__ " > < " __TIME__" >\n");
    CardServices(GetCardServicesInfo, &serv);
    if (serv.Revision != CS_RELEASE_CODE) {
        printk(KERN_NOTICE "lapcan_cs: Card Services release "
            "does not match!\n");
        return -1;
    }
    register_pccard_driver(&devInfo, &lapcanAttach, &lapcanDetach);
    error = register_chrdev(CAN_MAJOR, device_name, &lapcanFOps);
    /* Register driver for device */
    if (error<0) {
        DEBUG(1,"register_chrdev(%d, %s, %x) failed, error = %d\n",
            CAN_MAJOR, device_name, (int) &lapcanFOps, error);
        unregister_chrdev(CAN_MAJOR, device_name);
        return -1;
    }    
    return 0;
}

/*======================================================================*/
/* Run when driver is unloaded                                          */
/*======================================================================*/

void cleanup_module(void)
{
    unregister_pccard_driver(&devInfo);
    while (devList != NULL) {
        if (devList -> state & DEV_CONFIG)
            lapcanRelease((u_long)devList);
        lapcanDetach(devList);
    }
    unregister_chrdev(CAN_MAJOR, device_name);
}

/*======================================================================*/










